"""Bernoulli loss function implementation (log-loss)."""

# author: Benjamin Cross
# email: btcross26@yahoo.com
# created: 2019-08-26


import numpy as np

from .base_class import BaseLoss


class LogLoss(BaseLoss):
    """Log loss function class."""

    def __init__(self, eps: float = 1e-12):
        """
        Class initializer.

        Extends the BaseLoss class intializer.

        Parameter
        ---------
        eps: float (default = 1e-10)
            A small constant float to prevent log from returning negative infinity.
        """
        super().__init__()
        self._eps = eps

    def _loss(self, yt: np.ndarray, yp: np.ndarray) -> np.ndarray:
        """
        Calculate the per-observation loss as a function of `yt` and `yp`.

        Overrides BaseLoss._loss.
        """
        return -yt * np.log(yp + self._eps) - (1.0 - yt) * np.log(1.0 - yp + self._eps)

    def dldyp(self, yt: np.ndarray, yp: np.ndarray) -> np.ndarray:
        """
        Calculate the first derivative of the loss with respect to `yp`.

        Overrides BaseLoss.dldyp.
        """
        return -(yt / (yp + self._eps)) + (1.0 - yt) / (1.0 - yp + self._eps)

    def d2ldyp2(self, yt: np.ndarray, yp: np.ndarray) -> np.ndarray:
        """
        Calculate the second derivative of the loss with respect to `yp`.

        Overrides BaseLoss.d2ldyp2.
        """
        return (yt / (yp ** 2 + self._eps)) + (1.0 - yt) / ((1.0 - yp) ** 2 + self._eps)



"""
Unit tests for boosted_model class
"""

# author: Benjamin Cross
# email: btcross26@yahoo.com
# created: 2020-03-31

import logging
from unittest import mock

import numpy as np
import pytest

LOGGER = logging.getLogger(__name__)


# test init and that model is not fit after initialization
def test_initialize_and_bool(boosted_model_instance):
    """
    Test for basic initialization and __bool__ magic method, which returns whether
    the model has been fit or not
    """
    # GIVEN a BoostedModel instance with internal link, loss, and model_callbacks
    model = boosted_model_instance

    # WHEN checking model fit, False should be returned
    assert not model


# test compute link
def test_compute_link(boosted_model_instance):
    """
    Test that compute link is calling _link object attribute
    """
    # GIVEN a BoostedModel instance with identity link
    model = boosted_model_instance

    # WHEN the instance link is mocked and compute_link called
    # THEN the link object attribute should have been called once
    with mock.patch.object(model, "_link") as mock_link:
        y = np.linspace(-100, 200, 1001)
        model.compute_link(y)

        # assertions
        mock_link.assert_called_once()


# test compute loss
def test_compute_loss(boosted_model_instance):
    """
    Test that compute_loss is calling _loss object attribute
    """
    # GIVEN a BoostedModel instance with identity link
    model = boosted_model_instance

    # WHEN the instance link is mocked and compute_link called
    # THEN the link object attribute should have been called once
    with mock.patch.object(model, "_loss") as mock_loss:
        yp = np.linspace(-100, 200, 1001)
        yt = yp + np.random.randn(*yp.shape)
        model.compute_loss(yt, yp)

        # assertions
        mock_loss.assert_called_once()


# test compute gradients
def test_compute_gradients(boosted_model_instance):
    """
    Test that compute_loss is calling _loss object attribute
    """
    # GIVEN a BoostedModel instance with identity link
    model = boosted_model_instance

    # WHEN the instance link is mocked and compute_link called
    # THEN the link object attribute should have been called once
    with mock.patch.object(model, "_loss") as mock_loss, mock.patch.object(
        model, "_link"
    ) as mock_link:
        yp = np.linspace(-100, 200, 1001)
        yt = yp + np.random.randn(*yp.shape)
        model.compute_gradients(yt, yp)

        # assertions
        mock_loss.dldyp.assert_called_once()
        mock_link.dydeta.assert_called_once()


# test compute newton weights 1
def test_compute_newton_weights_1(boosted_model_instance):
    """
    Test that compute_newton_weights makes the right calls
    """
    # GIVEN a BoostedModel instance with identity link
    model = boosted_model_instance

    # WHEN the instance link is mocked and compute_newton_weights called
    # THEN the right link/loss calls should be made
    with mock.patch.object(model, "_loss") as mock_loss, mock.patch.object(
        model, "_link"
    ) as mock_link:
        yp = np.linspace(-100, 200, 1001)
        yt = yp + np.random.randn(*yp.shape)
        model.compute_newton_weights(yt, yp)

        # assertions
        mock_loss.d2ldyp2.assert_called_once()
        mock_link.dydeta.assert_called_once()
        mock_loss.dldyp.assert_called_once()
        mock_link.d2ydeta2.assert_called_once()


# test compute newton weights 2
def test_compute_newton_weights_2(boosted_model_instance):
    """
    Test that compute_newton_weights makes the right calls
    """
    # GIVEN a BoostedModel instance with identity link
    model = boosted_model_instance

    # WHEN the Identity/Least Squares link combo is called
    yp = np.linspace(-100, 200, 1001)
    yt = yp + np.random.randn(*yp.shape)
    weights = model.compute_newton_weights(yt, yp)

    # THEN all newton weights should be equal to 1.0
    np.all(weights == 1.0)


# test compute weights
@pytest.mark.parametrize(
    "weights_value",
    ["none", "newton", mock.MagicMock()],
    ids=["none", "newton", "callable"],
)
def test_compute_weights(boosted_model_instance, weights_value):
    """
    Test that compute_newton_weights makes the right calls
    """
    # GIVEN a BoostedModel instance with identity link
    model = boosted_model_instance

    # WHEN relevant attributes to compute_weights are mocked
    # THEN the right link/loss calls should be made
    with mock.patch.object(model, "weights", weights_value), mock.patch.object(
        model, "compute_newton_weights"
    ) as mock_cnw:
        yp = np.linspace(-100, 200, 1001)
        yt = yp + np.random.randn(*yp.shape)
        weights = model.compute_weights(yt, yp)

        # assertions
        if weights_value == "none":
            assert weights == 1.0
        elif weights_value == "newton":
            mock_cnw.assert_called_once()
        elif callable(weights_value):
            weights_value.assert_called_once()


# test compute weights exception
@pytest.mark.parametrize(
    "weights_value", [1, 2, None, "None", "Newton", "error"],
)
def test_compute_weights_exception(boosted_model_instance, weights_value):
    """
    Test that compute_weights throws exception at the right times
    """
    # GIVEN a BoostedModel instance with identity link
    model = boosted_model_instance

    # WHEN relevant attributes to compute_weights are mocked
    # THEN exception should be properly thrown with errant args
    with mock.patch.object(model, "weights", weights_value), mock.patch.object(
        model, "compute_newton_weights"
    ) as mock_cnw:
        yp = np.linspace(-100, 200, 1001)
        yt = yp + np.random.randn(*yp.shape)
        with pytest.raises(AttributeError) as excinfo:
            weights = model.compute_weights(yt, yp)
    assert (
        excinfo.value.args[0] == "attribute:<weights> should be 'none', 'newton', "
        "or a callable"
    )



"""BoostedLinearModel implementation for boosting models with linear coefficients."""

# author: Benjamin Cross
# email: btcross26@yahoo.com
# created: 2019-08-26


import warnings
from collections import OrderedDict
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np

from .boosted_model import BoostedModel
from .link_functions import BaseLink
from .loss_functions import BaseLoss
from .type_hints import LinearModel, ModelCallback, WeightsCallback


class BoostedLinearModel(BoostedModel):
    """BoostedLinearModel class implementation."""

    def __init__(
        self,
        link: BaseLink,
        loss: BaseLoss,
        model_callback: Callable[..., LinearModel],
        model_callback_kwargs: Optional[Dict[str, Any]] = None,
        weights: Union[str, WeightsCallback] = "none",
        alpha: float = 1.0,
        step_type: str = "default",
        step_decay_factor: float = 0.6,
        init_type: str = "mean",
        random_state: Optional[int] = None,
        validation_fraction: float = 0.0,
        validation_stratify: bool = False,
        validation_iter_stop: int = 10,
        tol: Optional[float] = None,
    ):
        """
        Class initializer.

        Parameters
        ----------
        link: BaseLink
            Link function to use in boosting iterations.

        loss: BaseLoss
            Loss function to use in boosting iterations.

        model_callback: Callable
            A callable that returns a model object that implements fit and predict
            methods. The model object that is returned must be a linear model that
            has coef_ and intercept_ attributes.

        model_callback_kwargs: dict, optional (default=None)
            A dictionary of keyword arguments to pass to `model_callback`.

        weights: Union[str, WeightsCallback]: str or Callable
            A string specificying the type of weights (one of "none" or "newton"), or a
            callable of the form 'lambda yt, yp: weights`, where yt and yp are the
            actual target and predicted target arrays. These weights are multiplied
            element-wise by the model gradients to produce the pseudo-residuals that are
            to be predicted at each model iteration.

        alpha: float (default=1.0)
            A parameter representing the intial trial learning rate. The learning rate that
            actually gets used at each iteration is dependent upon the value of `step_type`.

        step_type: str (default="decaying")
            One of "decaying", "constant", or "best". For "decaying", the initial model
            iteration will start with `alpha` as the learning parameter. If `alpha` times
            `step_decay_factor` results in a greater reduction in loss than `alpha`, then
            use the former. This is repeated until performance does not improve, and the
            final chosen rate will serve as the 'initial' rate for the next boosting
            iteration. For "constant", `alpha` will be used at every boosting iteration.
            Using "step_type" best implements the same process as "decaying", except the
            next boosting iteration will reset the learning rate back to `alpha` as the
            initial trial learning rate.

        step_decay_factor: float (default=0.48)
            The decaying factor to use with `step_type` "decaying" or "best".

        init_type: str (default="mean")
            The type of intial prediction to use. If "mean", then the initial link
            prediction (prior to boosting iterations) will be taken as the link of the mean
            of the non-transformed target. If "residuals" or "zero", the initial link
            prediction will be set to 0.0.

        random_state: int, optional (default=None)
            Random seed to be used for reproducability when random methods are used
            internally.

        validation_fraction: float (default=0.0)
            If 0.0, then no validation set will be used and training will be performed
            using the full training set. Otherwise, the fraction of observations to use
            as a holdout set for early stopping.

        validation_stratify: bool (default=False)
            If true, than stratify the validation holdout set based on the target. Useful
            when the target is binary.

        validation_iter_stop: int (default=10)
            Number of iterations to use for early stopping on the validation set. If the
            holdout loss is greater at the current iteration than `validation_iter_stop`
            iterations prior, then stop model fitting.

        tol: float, optional (default=None)
            Early stopping criteria based on training loss. If training loss fails to
            improve by at least `tol`, then stop training. If None, then training loss
            criteria is not checked to determine early stopping.
        """
        super().__init__(
            link,
            loss,
            model_callback,
            model_callback_kwargs,
            weights,
            alpha,
            step_type,
            step_decay_factor,
            init_type,
            random_state,
            validation_fraction,
            validation_stratify,
            validation_iter_stop,
            tol,
        )

        self.coef_: np.ndarray
        self.intercept_: float

    def boost(
        self,
        X: np.ndarray,
        yt: np.ndarray,
        yp: np.ndarray,
        eta_p: np.ndarray,
        model_callback: ModelCallback,
        model_callback_kwargs: Dict[str, Any],
        weights: Optional[np.ndarray] = None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Boost by one model iteration.

        Creates a model using the `model_callback` callable with `model_callback_kwargs`,
        then fits this model to the pseudo-residuals. The learning rate is determined
        according to the chosen method, and current predictions are updated and returned
        such that stopping criteria can be evaluated and boosting continued. The
        fitted model resulting from the iteration is appended to the underlying model
        ensemble.

        Parameters
        ----------
        X: np.ndarray
            The model matrix. If `init_type` was set as "residuals", then the model
            scores for which to calculate residuals should form the last column of the
            input matrix.

        yt: np.ndarray
            Observed target values.

        yp: np.ndarray
            Predicted target values.

        eta_p: np.ndarray
            The current link predictions corresponding to the model iteration. This can
            be found by transforming `yp`, but passing this as an argument avoids
            duplicate computations and improves performance.

        model_callback: Callable
            A callable that returns a model object that implements fit and predict methods.

        model_callback_kwargs: dict, optional (default=None)
            A dictionary of keyword arguments to pass to `model_callback`.

        weights: np.ndarray, optional (default=None)
            Sample weights (by observation) to use for fitting. Should be positive.
            Observations with higher weights will affect the model fit more. If 'None',
            then all weights will be equal (1.0).

        Returns
        -------
        yp_next, eta_p_next: tuple(np.ndarray, np.ndarray)
            A tuple of the updated target predictions and target prediction links.
        """
        yp_next, eta_p_next = super().boost(
            X, yt, yp, eta_p, model_callback, model_callback_kwargs, weights
        )
        model, lr = self._model_list[-1]
        self.coef_ += lr * model.coef_  # type: ignore
        self.intercept_ += lr * model.intercept_  # type: ignore
        return yp_next, eta_p_next

    def decision_function(
        self, X: np.ndarray, model_index: Optional[int] = None
    ) -> np.ndarray:
        """
        Get the link of computed model predictions.

        Parameters
        ----------
        X: np.ndarray
            The model matrix. If `init_type` was set as "residuals", then the model
            scores for which to calculate residuals should form the last column of the
            input matrix.

        model_index: int, optional (default=None)
            If None, then return the full model prediction as the sum of all models in
            the ensemble plus the initial model prediction. If an int, then return only
            the predictions from models up to `model_index` (i.e., [:model_index]).

        Returns
        -------
        np.ndarray
            The link of the computed model predictions.
        """
        if model_index is not None:
            warnings.warn(
                "arg:model_index is ignored for BoostedLinearModel.decision_function"
            )
        if self.get_iterations() == 0:
            eta_p = self._model_init.predict(X)
        else:
            eta_p = self.intercept_ + X[:, : self._msi].dot(self.coef_)
        return eta_p

    def get_coefficient_order(self, scale: Optional[np.ndarray] = None) -> List[int]:
        """
        Get the order that coefficients were selected for the model.

        In the case that multiple coefficients were selected for the first time at the
        same model iteration, the "larger" coefficient will be considered to have been
        selected first. The `scale` argument can be used to standardize coefficients
        if models were fitted in a manner such that coefficients were not standardized.

        Parameters
        ----------
        scale: np.ndarray, optional (default=None)
            Vector to scale the coefficients in the ordering calculation. If None, then
            coefficients are not scaled (or alternatively, all coefficients are scaled
            by a factor of 1.0).

        Returns
        -------
        coefficient_order: List[int]
            A list of zero-based indexes specifying the order that coefficients entered
            the boosted model.
        """
        scale = 1.0 if scale is None else scale
        coef_order_dict = OrderedDict()  # type: Dict[int, None]
        for model, _ in self._model_list:
            coefs = model.coef_ * scale  # type: ignore
            nc = (coefs != 0.0).sum()
            order = np.argsort(np.abs(coefs))[::-1].tolist()
            coef_order_dict.update(OrderedDict.fromkeys(order[:nc]))
        return list(coef_order_dict.keys())

    def get_coefficient_history(self, scale: Optional[np.ndarray] = None) -> np.ndarray:
        """
        Get the history of coefficient values.

        Returns a matrix of coefficients, where each row is representative of the
        model coefficients at a specific boosting iteration (i.e., row 1 is after the
        first round of boosting, etc.)

        Parameters
        ----------
        scale: np.ndarray, optional (default=None)
            Vector to scale the coefficients in the history calculation. If None, then
            coefficients are not scaled (or alternatively, all coefficients are scaled
            by a factor of 1.0).

        Returns
        -------
        coefficient_history: np.ndarray [n_boosting_iterations + 1, n_coefs]
        """
        scale = 1.0 if scale is None else scale.reshape((1, -1))
        if self._is_fit:
            coef_history = list()
            for i, (model, lr) in enumerate(self._model_list):
                coef = model.coef_ * lr  # type: ignore
                if i == 0:
                    coef_history.append(coef)
                    continue
                coef = coef + coef_history[i - 1]
                coef_history.append(coef)
            coef_matrix = np.vstack([np.zeros_like(coef), coef_history])
            coef_matrix *= scale
            return coef_matrix
        else:
            raise AttributeError("model has not yet been fit")

    def get_prediction_var_history(
        self, X: np.ndarray, groups: Optional[List[int]] = None
    ) -> np.ndarray:
        """
        Get the history of prediction variance on the model matrix X.

        Returns a matrix of prediction variances at each round of boosting for each
        specified group of model coefficients. If `groups` is None, then each
        coefficient is considered separately as its own group.

        Parameters
        ----------
        X: np.ndarray
            The model matrix. If `init_type` was set as "residuals", then the model
            scores for which to calculate residuals should form the last column of the
            input matrix.

        groups: List[int], optional (default=None)
            A list of indices representing coefficient groups. Indices for groups should
            start at zero and be sequenced in order to the number of groups minus one.
            If None, then each feature is its own group.

        Returns
        -------
        pred_var: np.ndarray [n_boosting_iterations, n_groups]
        """
        coef_history = self.get_coefficient_history()
        pred_vars = np.zeros_like(coef_history)
        Xc = X[:, : self._msi] - X[:, : self._msi].mean(axis=0).reshape((1, -1))

        for i, coef in enumerate(coef_history):
            preds = Xc * coef.reshape((1, -1))
            pred_vars[i, :] = preds.var(axis=0)

        if groups is not None:
            groups = np.array(groups)
            max_ind = np.max(groups)
            group_vars = np.zeros((pred_vars.shape[0], max_ind + 1))
            for i in range(max_ind + 1):
                col_index = np.nonzero(groups == i)[0].tolist()
                group_vars[:, i] = np.sum(
                    (pred_vars[:, col_index] * coef_history[:, col_index]) ** 2, axis=1
                )
            pred_vars = group_vars

        return pred_vars

    def initialize_model(
        self, X: np.ndarray, yt: np.ndarray, weights: Optional[np.ndarray] = None
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Initialize the boosted model.

        This is called internally within the fit function. If manual boosting is being
        performed using the 'boost' method, then this method should be called before
        beginning the manual boosting procedure.

        Parameters
        ----------
        X: np.ndarray [n_samples, n_features]
            The model matrix. If `init_type` was set as "residuals", then the model
            scores for which to calculate residuals should form the last column of the
            input matrix.

        yt: np.ndarray
            Observed target values.

        weights: np.ndarray, optional (default=None)
            Sample weights (by observation) to use for fitting. Should be positive.
            Observations with higher weights will affect the model fit more. If 'None',
            then all weights will be equal (1.0).

        Returns
        -------
        yp, eta_p: tuple(np.ndarray, np.ndarray)
            The initial target predictions and link of target prediction arrays.
        """
        yp, eta_p = super().initialize_model(X, yt, weights)

        if self.get_iterations() == 0:
            self.coef_ = np.zeros(X[:, : self._msi].shape[1])
            self.intercept_ = self._model_init._value

        return yp, eta_p



"""
Helper functions to approximate gradients.

Uses central differences for use in testing and numerical derivatives
"""

# author: Benjamin Cross
# email: btcross26@yahoo.com
# created: 2019-12-16


from typing import Callable

import numpy as np


def d1_central_difference(
    func: Callable[[np.ndarray], np.ndarray], y: np.ndarray, h: float = 1e-8
) -> np.ndarray:
    """
    Approximate the first derivative of func at y.

    Uses a first-order central difference with spacing `h` (error term O(h^2))

    Parameters
    ----------
    func: callable
        Function that takes y as an argument and returns a value

    y: np.ndarray
        Locations to calculate the value of func

    h: float
        Spacing to use in calculating central difference. This value should be
        greater than zero.

    Returns
    -------
    np.ndarray
        The approximated values of the first derivative
    """
    value = (func(y + h) - func(y - h)) / (2.0 * h)
    return value


def d2_central_difference(
    func: Callable[[np.ndarray], np.ndarray], y: np.ndarray, h: float = 1e-8
) -> np.ndarray:
    """
    Approximate the first derivative of func at y.

    Uses a first-order central difference with spacing `h` (error term O(h^2))

    Parameters
    ----------
    func: callable
        Function that takes y as an argument and returns a value

    y: np.ndarray
        Locations to calculate the value of func

    h: float
        Spacing to use in calculating central difference. This value should be
        greater than zero.

    Returns
    -------
    np.ndarray
        The approximated values of the first derivative
    """
    value = (func(y + h) - 2.0 * func(y) + func(y - h)) / (h ** 2)
    return value



"""genestboost.loss_functions module."""

# author: Benjamin Cross
# email: btcross26@yahoo.com
# created: 2019-08-26


from .absolute_value import AbsoluteLoss
from .base_class import BaseLoss
from .beta_loss import BetaLoss, LeakyBetaLoss
from .least_squares import LeastSquaresLoss
from .log_cosh import LogCoshLoss
from .log_loss import LogLoss
from .poisson import PoissonLoss
from .quantile import QuantileLoss
from .quasilog_loss import QuasiLogLoss
from .students_t import StudentTLoss

__all__ = [
    "AbsoluteLoss",
    "BaseLoss",
    "BetaLoss",
    "LeakyBetaLoss",
    "LeastSquaresLoss",
    "LogCoshLoss",
    "LogLoss",
    "PoissonLoss",
    "QuantileLoss",
    "QuasiLogLoss",
    "StudentTLoss",
]



"""
Meta-fixtures for unit tests
"""

# author: Benjamin Cross
# email: btcross26@yahoo.com
# created: 2020-04-07


import pytest

from genestboost import BoostedModel
from genestboost.link_functions import IdentityLink
from genestboost.loss_functions import LeastSquaresLoss
from genestboost.weak_learners import SimplePLS


# fixture for generic boosted model object
@pytest.fixture(scope="function")
def boosted_model_instance():
    # setup
    model = BoostedModel(
        link=IdentityLink(),
        loss=LeastSquaresLoss(),
        model_callback=SimplePLS,
        model_callback_kwargs={"max_vars": 3, "filter_threshold": None},
    )
    # yield objects
    yield model

    # teardown
    del model



"""
List of test params to import into link_function_tests.py
"""

# author: Benjamin Cross
# email: btcross26@yahoo.com
# created: 2019-12-16

import numpy as np

from genestboost.link_functions import (
    CLogLogLink,
    CubeRootLink,
    IdentityLink,
    LogitLink,
    LogLink,
    Logp1Link,
    PowerLink,
    ReciprocalLink,
    SqrtLink,
)

# setup link test params to loop through tests
LINK_TESTS = list()

# identity test params
LINK_TESTS.append(
    (
        "identity_test",
        IdentityLink(),
        np.array([[-5.0, -5.0], [0.0, 0.0], [10.0, 10.0]]),
        np.linspace(-50, 50, 101),
        (0.005, 0.0001),
    )
)

# logit test params
LINK_TESTS.append(
    (
        "logit_test",
        LogitLink(),
        np.array([[0.5, 0.0], [0.25, -np.log(3)], [0.9, np.log(9)]]),
        np.linspace(0.01, 0.99, 99),
        (0.001, 0.0001),
    )
)

# cloglog test params
LINK_TESTS.append(
    (
        "cloglog_test",
        CLogLogLink(),
        np.array(
            [
                [0.5, np.log(-np.log(0.5))],
                [0.25, np.log(-np.log(0.75))],
                [0.9, np.log(-np.log(0.1))],
            ]
        ),
        np.linspace(0.01, 0.99, 99),
        (0.001, 0.0001),
    )
)

# log link test params
LINK_TESTS.append(
    (
        "log_test",
        LogLink(),
        np.array([[1.0, 0.0], [np.exp(1), 1.0], [np.exp(10), 10.0]]),
        np.logspace(0.0, 10.0, 111, base=np.exp(1)),
        (0.001, 0.001),
    )
)

# logp1 link test params
LINK_TESTS.append(
    (
        "logp1_test",
        Logp1Link(),
        np.array([[0.0, 0.0], [np.exp(1) - 1.0, 1.0], [np.exp(10) - 1.0, 10.0]]),
        np.logspace(0.0, 10.0, 111, base=np.exp(1)),
        (0.001, 0.001),
    )
)

# power link (squared + 1) test params
LINK_TESTS.append(
    (
        "power_base_test",
        PowerLink(power=2, summand=1.0),
        np.array([[0.0, 1.0], [2.0, 9.0], [10.0, 121.0]]),
        np.linspace(0.0, 100.0, 101),
        (0.005, 0.001),
    )
)

# sqrt link test params
LINK_TESTS.append(
    (
        "sqrt_link_test",
        SqrtLink(),
        np.array([[0.01, 0.1], [4.0, 2.0], [100.0, 10.0]]),
        np.linspace(0.01, 10.0, 101),
        (0.001, 0.001),
    )
)

# power link (squared + 1) test params
LINK_TESTS.append(
    (
        "cuberoot_link_test",
        CubeRootLink(),
        np.array([[0.001, 0.1], [8.0, 2.0], [1000.0, 10.0]]),
        np.linspace(0.01, 10.0, 101),
        (0.001, 0.001),
    )
)

# power link (squared + 1) test params
LINK_TESTS.append(
    (
        "reciprocal_link_test",
        ReciprocalLink(),
        np.array([[1.0, 1.0], [2.0, 0.5], [10.0, 0.1]]),
        np.linspace(0.01, 10.0, 101),
        (0.001, 0.001),
    )
)



"""Link functions related to log."""

# author: Benjamin Cross
# email: btcross26@yahoo.com
# created: 2019-08-26


import numpy as np

from .base_class import BaseLink


class LogLink(BaseLink):
    """Implementation of the log link function."""

    def __init__(self, summand: float = 0.0):
        """
        Class initializer.

        Extends the BaseLink class intializer.

        Parameters
        ----------
        summand: float
            Summand of the log in the link implementation - i.e.,
            link = log(y + summand).
        """
        super().__init__()
        self.summand_ = summand

    def _link(self, y: np.ndarray) -> np.ndarray:
        """
        Get the link, eta, as a function of y.

        Overrides BaseLink._link.
        """
        return np.log(y + self.summand_)

    def _inverse_link(self, eta: np.ndarray) -> np.ndarray:
        """
        Get the target, y, as a function of the link, `eta`.

        Overrides BaseLink._inverse_link.
        """
        return np.exp(eta) - self.summand_

    def dydeta(self, y: np.ndarray) -> np.ndarray:
        """
        Get the derivative of `y` with respect to the link as a function of y.

        Overrides BaseLink.dydeta.
        """
        return y + self.summand_

    def d2ydeta2(self, y: np.ndarray) -> np.ndarray:
        """
        Get the second derivative of `y` with respect to the link as a function of y.

        Overrides BaseLink.d2ydeta2.
        """
        return y + self.summand_


class Logp1Link(LogLink):
    """Log plus 1 link function implementation."""

    def __init__(self) -> None:
        """Class initializer extends LogLink initializer, setting summand equal to 1.0."""
        super().__init__(summand=1.0)



"""QuasiLogLoss function implementation."""

# author: Benjamin Cross
# email: btcross26@yahoo.com
# created: 2019-10-16

from typing import Callable

import numpy as np

from .base_class import BaseLoss


class QuasiLogLoss(BaseLoss):
    """QuasiLogLoss loss function class."""

    def __init__(
        self,
        vt_callback: Callable[[np.ndarray], np.ndarray],
        d0_n: int = 100,
        d2_eps: float = 1e-12,
    ):
        """
        Class initializer.

        Extends BaseLoss.__init__.

        Parameters
        ----------
        vt_callback: callable(yp) -> np.ndarray
            Callable that takes the predicted value and calculates the denominator of
            the quasilog loss integral for each observation.

        d0_n: int
            The number of points to use for integration.

        d2_eps: float
            A small float used to calculate the second derivative via the central
            difference formula - 2*d2_eps the interval used to calculate the derivative
            from first derivative values.
        """
        super().__init__()
        self._vt_callback = vt_callback
        self._d0_n = d0_n + (d0_n % 2)
        self._d2_eps = d2_eps

    def _loss(self, yt: np.ndarray, yp: np.ndarray) -> np.ndarray:
        """
        Calculate the per-observation loss as a function of `yt` and `yp`.

        Overrides BaseLoss._loss.
        """
        # use Simpson's rule to numerically integrate
        dims = tuple([1] * yp.ndim + [-1])
        x = np.linspace(0.0, 1.0, self._d0_n + 1).reshape(dims)
        iwts = np.hstack(
            [[1.0], np.tile([4.0, 2.0], (self._d0_n - 2) // 2), [4.0, 1.0]]
        ).reshape(dims)
        ipts = np.expand_dims(yp, axis=-1) + np.expand_dims(yt - yp, axis=-1) * x
        values = np.sum(self.dldyp(np.expand_dims(yt, axis=-1), ipts) * iwts, axis=-1)
        values = values * (yp - yt) / (3.0 * self._d0_n)
        return values

    def dldyp(self, yt: np.ndarray, yp: np.ndarray) -> np.ndarray:
        """
        Calculate the first derivative of the loss with respect to `yp`.

        Overrides BaseLoss.dldyp.
        """
        return -(yt - yp) / self._vt_callback(yp)

    def d2ldyp2(self, yt: np.ndarray, yp: np.ndarray) -> np.ndarray:
        """
        Calculate the second derivative of the loss with respect to `yp`.

        Overrides BaseLoss.d2ldyp2.
        """
        v1 = self.dldyp(yt, yp - self._d2_eps)
        v2 = self.dldyp(yt, yp + self._d2_eps)
        return (v2 - v1) / (2.0 * self._d2_eps)



"""Link functions module - contains link functions for boosting."""

# author: Benjamin Cross
# email: btcross26@yahoo.com
# created: 2019-08-26


from .base_class import BaseLink
from .cloglog import CLogLogLink
from .identity import IdentityLink
from .log_links import LogLink, Logp1Link
from .logit import LogitLink
from .power_links import CubeRootLink, PowerLink, ReciprocalLink, SqrtLink

__all__ = [
    "BaseLink",
    "CLogLogLink",
    "IdentityLink",
    "LogLink",
    "Logp1Link",
    "LogitLink",
    "CubeRootLink",
    "PowerLink",
    "ReciprocalLink",
    "SqrtLink",
]



"""Link function abstract base class."""

# author: Benjamin Cross
# email: btcross26@yahoo.com
# created: 2019-08-26


from abc import ABC, abstractmethod

import numpy as np


class BaseLink(ABC):
    """Abstract base class for link functions."""

    def __call__(self, y: np.ndarray, inverse: bool = False) -> np.ndarray:
        """
        Call the instance object to calculate the link function.

        If `inverse` is True (default), get the link, eta, as a function of `y`.
        If `inverse` is False, then get y as a function of the link. In this case,
        one should pass the link function eta as the argument `y`.
        """
        if inverse:
            return self._inverse_link(y)
        return self._link(y)

    @abstractmethod
    def _link(self, y: np.ndarray) -> np.ndarray:
        """Get the link, eta, as a function of y."""
        ...

    @abstractmethod
    def _inverse_link(self, eta: np.ndarray) -> np.ndarray:
        """Get the target, y, as a function of the link, `eta`."""
        ...

    @abstractmethod
    def dydeta(self, y: np.ndarray) -> np.ndarray:
        """Get the derivative of `y` with respect to the link as a function of y."""
        ...

    @abstractmethod
    def d2ydeta2(self, y: np.ndarray) -> np.ndarray:
        """Get the second derivative of `y` with respect to the link as a function of y."""
        ...



"""genestboost top-level module."""

# author: Benjamin Cross
# email: btcross26@yahoo.com
# created: 2019-08-28

from .boosted_linear_model import BoostedLinearModel
from .boosted_model import BoostedModel
from .model_data_sets import ModelDataSets

__version__ = "0.3.1"

__all__ = ["BoostedLinearModel", "BoostedModel", "ModelDataSets", "__version__"]



"""Implementation for SimplePLS, a partial-PLS component regression model."""

# author: Benjamin Cross
# email: btcross26@yahoo.com
# created: 2019-08-28


import heapq
from typing import List, Optional, Tuple

import numpy as np

from ..type_hints import Model


class SimplePLS:
    """Implementation for SimplePLS, a partial-PLS component regression model."""

    def __init__(
        self, max_vars: Optional[int] = 1, filter_threshold: Optional[float] = None
    ):
        """
        Class SimplePLS initializer.

        Parameters
        ----------
        max_vars: int, optional (default=1)
            The maximum number of variables to use in the regression. The default value
            is 1, which is the special case of simple one-variable least squares
            regression. If None, then max_vars will be set to the number of variables
            in the X model matrix during the fitting process.

        filter_threshold: float, optional (default=None)
            The correlation filter threshold to use. If the ratio of the absolute value
            of the correlation coefficient for a predictor to the absolute value of the
            maximum correlation coefficient of all predictors is less than the filter
            threshold value, then the predictor will be excluded from the regression.
            The default value is None, in which case the filter threshold will be set
            equal to 0.0.

        Attributes
        ----------
        coef_ : numpy.ndarray, shape (n_predictors,)
            The estimated coefficients for the regression problem

        intercept_ : float
            The estimated intercept (bias) for the regression problem
        """
        # initialized attributes
        self._max_vars = max_vars
        self._filter_threshold = 0.0 if filter_threshold is None else filter_threshold

        # public attributes initialized during class usage
        self.coef_: np.ndarray
        self.intercept_: float

        # private attributes initialized during class usage
        self._X_means: np.ndarray
        self._X_std: np.ndarray
        self._y_mean: float
        self._y_std: float
        self._multiplier: float
        self._pls_intercept: float

    def fit(
        self, X: np.ndarray, y: np.ndarray, weights: Optional[np.ndarray] = None
    ) -> Model:
        """
        Fit a linear regression according to the specified initializer arguments.

        The fit process will result in several instance attributes being populated,
        including the public attributes housing the coefficients and intercept.

        Parameters
        ----------
        X: numpy.ndarray, shape (n_samples, n_predictors)
            The input model matrix. Should be of dtype float.

        y: numpy.ndarray, shape (n_samples, )
            The target vector

        weights: numpy.ndarray, optional (default=None)
            A value of weights to use for the linear regression fit. The default value
            is None, which results in equal weighting for all observations. If a value
            is provided, it should have the same dimensions as the target vector, y.
            The weights are applied to the correlation calculation only. They are not
            used during standardization of the X and y fitted arrays at model
            initialization.

        Returns
        -------
        model: SimplePLS
            Instance of self.
        """
        # initialize model
        Xs, ys = self._initialize_model(X, y)

        # calculate initial model coefficients
        weights = np.ones(y.shape[0]) if weights is None else weights
        coefs = np.sum(Xs * (ys * weights).reshape((-1, 1)), axis=0) / np.sum(weights)

        # mask/filter coefficients
        coefs = self._mask_coefs(coefs)

        # get regression coefficients
        self._multiplier = self._get_coef_multiplier(coefs, Xs, ys)
        coefs *= self._multiplier
        coefs *= self._y_std / self._X_std[0]
        self.coef_ = coefs

        # get intercept
        self.intercept_ = (
            self._y_mean
            - np.sum(coefs * self._X_means[0])
            - self._pls_intercept * self._y_std
        )

        return self

    def predict(self, X: np.ndarray) -> np.ndarray:
        """
        Compute model predictions for the given model matrix, X.

        Parameters
        ----------
        X: numpy.ndarray, shape (n_samples, n_predictors)
             The input model matrix. Should be of dtype float. The number of columns
             should be the same number of columns as the X argument that was used to
             fit the model.

        Returns
        -------
        numpy.ndarray, shape (n_samples, )
            A vector of predictions of dtype float
        """
        return self.intercept_ + X.dot(self.coef_)

    def _get_coef_multiplier(
        self, coefs: np.ndarray, Xs: np.ndarray, ys: np.ndarray
    ) -> np.ndarray:
        """
        Get the coefficient multiplier that is applied to the coefficients (private).

        In the case of single variable regression, this multiplier will be equal to
        1.0. For multiple variables, A regression is applied to the summation of each
        selected standardized variable multiplied by its correlation coefficient. The
        final regression coefficient for a variable is then equal to its correlation
        coefficient multiplied by the multiplier returned by this method. The
        multiplier that is returned by this method is also stored in the private
        instance attribute, _multiplier.

        Parameters
        ----------
        coefs: numpy.ndarray, shape (n_predictors,)
            A vector of coefficients. For selected variables, the coefficients will be
            equal to the individual correlation coefficients of the variables with the
            target. For variables that are not selected, the coefficient values are
            equal to zero.

        Xs: numpy.ndarray, shape (n_samples, n_predictors)
            The standardized version of model matrix, X, that is being fitted

        ys: numpy.ndarray, shape (n_samples,)
            The standardized version of target vector, y, that is being fitted

        Returns
        -------
        float
            The multiplier that is applied to the selected variable correlation
            coefficients to get the final regression model
        """
        # initialize values
        self._pls_intercept = 0.0
        multiplier = 1.0
        n_coefs = np.sum(coefs != 0.0)

        # regress on weighted (by correlation coefficient) sum if max_vars is greater
        # than 1
        if n_coefs != 1:
            x_pls = Xs.dot(coefs)
            x_pls_mean = x_pls.mean()
            x_pls_std = x_pls.std()
            x_pls_std = np.where(x_pls_std == 0.0, 1.0, x_pls_std)
            x_pls_scaled = (x_pls - x_pls_mean) / x_pls_std
            alpha = np.mean(x_pls_scaled * ys)
            self._pls_intercept = alpha * x_pls_mean / x_pls_std
            multiplier = alpha / x_pls_std

        return multiplier

    def _mask_coefs(self, coefs: np.ndarray) -> np.ndarray:
        """
        Apply initializer arguments max_vars and filter_threshold (private).

        If max_vars is 1 or filter_threshold is 1.0, then only the variable
        that is most correlated with the predictor will be selected. If max_vars is
        greater than or equal to the number of predictors and filter_threshold is
        <= 0.0, then all variables are selected. Otherwise, max_vars will be selected
        and paired down according to the value of filter_threshold in the general case.

        Parameters
        ----------
        coefs: numpy.ndarray, shape (n_predictors,)
            A vector of coefficients calculated during the fitting process. The values
            of the coefficients in this case are equal to the correlation coefficients
            of the predictors with the target vector.

        Returns
        -------
        numpy.ndarray
            The vector of coefficients after max_vars have been selected and paired
            down according to the specified filter_threshold.
        """
        # initialize values
        coefs_abs = np.abs(coefs)
        max_index = np.argmax(coefs_abs)

        # edge case where there will only be one var
        if self._max_vars == 1 or self._filter_threshold >= 1.0:
            coef_mask = np.zeros(coefs.shape[0])
            coef_mask[max_index] = 1.0
            return coefs * coef_mask

        # edge case where all vars used
        num_vars = (
            coefs.shape[0]
            if self._max_vars is None
            else min(self._max_vars, coefs.shape[0])
        )
        if num_vars == coefs.shape[0] and (
            self._filter_threshold <= 0.0 or self._filter_threshold is None
        ):
            return coefs

        # apply correlation filter
        max_value = coefs_abs[max_index]
        rel_coefs = coefs_abs / max_value
        if self._filter_threshold is not None:
            coef_mask = 1.0 * (rel_coefs >= self._filter_threshold)

        # apply max_vars intermediate case if specified
        if coef_mask.sum() > num_vars:
            heap_index = list()  # type: List[Tuple[float, int]]
            for i in np.nonzero(coef_mask == 1)[0]:
                value = rel_coefs[i]
                if len(heap_index) < num_vars:
                    heapq.heappush(heap_index, (value, i))
                else:
                    min_heap_tuple = heapq.heappop(heap_index)
                    if value > min_heap_tuple[0]:
                        heapq.heappush(heap_index, (value, i))
                    else:
                        heapq.heappush(heap_index, min_heap_tuple)
            mask_index = list(map(lambda x: x[1], heap_index))
            coef_mask = np.zeros(coefs.shape[0])
            coef_mask[mask_index] = 1.0

        return coef_mask * coefs

    def _initialize_model(self, X: np.ndarray, y: np.ndarray) -> np.ndarray:
        """
        Initialize the model by standardizing the X and y fitting arrays (private).

        Parameters
        ----------
        X: numpy.ndarray, shape (n_samples, n_predictors)
            The input model matrix. The number of columns should be the same number of
            columns as the X argument that was used to fit the model.

        y: numpy.ndarray, shape (n_samples, )
            The target vector

        Returns
        -------
        Xs: numpy.ndarray, shape (n_samples, n_predictors)
            The standardized version of model matrix, X, that is being fitted

        ys: numpy.ndarray, shape (n_samples, )
            The standardized version of target vector, y, that is being fitted
        """
        self._X_means = X.mean(axis=0, keepdims=True)
        self._X_std = X.std(axis=0, keepdims=True)
        self._X_std = np.where(self._X_std == 0, 1.0, self._X_std)
        self._y_mean = y.mean()
        self._y_std = y.std()
        self._y_std = 1.0 if self._y_std == 0.0 else self._y_std
        return (X - self._X_means) / self._X_std, (y - self._y_mean) / self._y_std



"""Least squares loss function implementation."""

# author: Benjamin Cross
# email: btcross26@yahoo.com
# created: 2019-11-03


from typing import Callable

import numpy as np
from scipy.optimize import bisect
from scipy.special import beta as beta_function
from scipy.special import betainc

from . import BaseLoss


class BetaLoss(BaseLoss):
    """BetaLoss class implementation."""

    def __init__(self, alpha: float, beta: float, eps: float = 1e-10):
        """
        Class initializer.

        Extends BaseLoss.__init__.

        Parameters
        ----------
        alpha: float
            Value of alpha in beta loss denominator, mu ** (1 - alpha)

        beta: float
            Value of beta in beta loss denominator, (1 - mu) ** (1 - beta)

        eps: float
            Small value to use in log calculations to avoid numerical error
        """
        super().__init__()
        self.alpha = alpha
        self.beta = beta
        self.eps = eps
        self.scale = 1.0 / beta_function(alpha, beta)
        self._vt_callback = self.beta_callback(alpha, beta, eps)

    @staticmethod
    def beta_callback(
        alpha: float, beta: float, eps: float = 1e-10
    ) -> Callable[[np.ndarray], np.ndarray]:
        """
        Compute the beta callback function for quasi-deviance.

        Parameters
        ----------
        alpha: float
            Value of alpha in beta loss denominator, mu ** (1 - alpha)

        beta: float
            Value of beta in beta loss denominator, (1 - mu) ** (1 - beta)

        eps: float
            Small value to use in log calculations to avoid numerical error

        Returns
        -------
        callable
            A callable that takes an np.ndarray and returns an np.ndarray after
            calculating the beta quasi-deviance denominator
        """
        scale = 1.0 / beta_function(alpha, beta)

        def vt_callback(yp: np.ndarray) -> np.ndarray:
            return np.exp(
                -np.log(scale)
                + (1.0 - alpha) * np.log(yp + eps)
                + (1.0 - beta) * np.log(1.0 - yp + eps)
            )

        return vt_callback

    def _loss(self, yt: np.ndarray, yp: np.ndarray) -> np.ndarray:
        """
        Calculate the per-observation loss as a function of `yt` and `yp`.

        Overrides BaseLoss._loss.
        """
        c1 = betainc(self.alpha, self.beta, yt) - betainc(self.alpha, self.beta, yp)
        c2 = betainc(self.alpha + 1.0, self.beta, yt) - betainc(
            self.alpha + 1.0, self.beta, yp
        )
        b1 = beta_function(self.alpha, self.beta)
        b2 = beta_function(self.alpha + 1, self.beta)
        return (yt * c1 * b1 - c2 * b2) * self.scale

    def dldyp(self, yt: np.ndarray, yp: np.ndarray) -> np.ndarray:
        """
        Calculate the first derivative of the loss with respect to `yp`.

        Overrides BaseLoss.dldyp.
        """
        return -(yt - yp) / self._vt_callback(yp)

    def d2ldyp2(self, yt: np.ndarray, yp: np.ndarray) -> np.ndarray:
        """
        Calculate the second derivative of the loss with respect to `yp`.

        Overrides BaseLoss.d2ldyp2.
        """
        callback_values = self._vt_callback(yp)
        d_left = 1.0 / callback_values
        du = (1.0 - self.alpha) / (yp + self.eps) - (1.0 - self.beta) / (
            1.0 - yp + self.eps
        )
        d_right = du * (yt - yp) / callback_values
        return d_left + d_right


class LeakyBetaLoss(BetaLoss):
    """Class implementation of LeakyBetaLoss loss function."""

    def __init__(
        self,
        alpha: float,
        beta: float,
        gamma: float = 1.0,
        eps: float = 1e-10,
        xtol: float = 1e-8,
    ):
        """
        Class initializer.

        Extends BetaLoss.__init__.

        Parameters
        ----------
        alpha: float
            Passed as the alpha argument to the BetaLoss initializer

        beta: float
            Passed as the beta argument to the BetaLoss initializer

        eps: float, optional
            Passed as the eps argument to the BetaLoss initializer (the default value
            is 1e-10)

        gamma: float, optional
            A float in the range (0.0, 1.0] specifying ... (the default value is 1.0)

        xtol: float, optional
            Passed as the xtol argument to scipy.optimize.bisect (the default value is
            1e-8)
        """
        self.gamma = gamma
        super().__init__(alpha=alpha, beta=beta, eps=eps)

        # find leaky point values
        self.rL = self.alpha / (self.alpha + self.beta)  # left x
        self.vL = super().dldyp(0.0, self.rL)  # left slope
        self.rR = 1.0 - self.rL  # right x
        self.vR = super().dldyp(1.0, 1.0 - self.rR)  # right slope

        # leaky slopes
        self.mL = self.gamma * self.vL
        self.mR = self.gamma * self.vR

        # transition pts
        floss = super().dldyp
        self.xL = bisect(
            lambda x: floss(0.0, x) - self.mL, self.rL, 1.0 - 1e-8, xtol=xtol
        )
        self.yL = super()._loss(0.0, self.xL)
        self.xR = bisect(
            lambda x: floss(1.0, x) - self.mR, 1e-8, 1.0 - self.rR, xtol=xtol
        )
        self.yR = super()._loss(1.0, self.xR)

    def _loss(self, yt: np.ndarray, yp: np.ndarray) -> np.ndarray:
        # calculate loss function values from regular betaloss
        values = super()._loss(yt, yp)

        # modify left shelf
        values = np.where(
            yt - yp < -self.xL, self.yL - self.mL * (-yp + self.xL), values
        )

        # modify right shelf
        values = np.where(
            yt - yp > 1.0 - self.xR,
            self.yR - self.mR * (yt - yp - 1.0 + self.xR),
            values,
        )

        return values

    def dldyp(self, yt: np.ndarray, yp: np.ndarray) -> np.ndarray:
        """
        Calculate the first derivative of the loss with respect to `yp`.

        Overrides BaseLoss.dldyp.
        """
        # calculate loss gradient values from regular betaloss
        values = super().dldyp(yt, yp)

        # modify left shelf
        values = np.where(yt - yp < -self.xL, self.mL, values)

        # modify right shelf
        values = np.where(yt - yp > 1.0 - self.xR, self.mR, values)

        return values

    def d2ldyp2(self, yt: np.ndarray, yp: np.ndarray) -> np.ndarray:
        """
        Calculate the second derivative of the loss with respect to `yp`.

        Overrides BaseLoss.d2ldyp2.
        """
        # calculate loss gradient values from regular betaloss
        values = super().d2ldyp2(yt, yp)

        # modify shelves with 0.0 second derivative
        values = np.where((yt - yp < -self.xL) | (yt - yp > 1.0 - self.xR), 0.0, values)

        return values



"""
Custom type hints for use in various places of module
"""

# Author: Benjamin Cross
# email: btcross26@yahoo.com
# created: 2019-11-15


from __future__ import annotations

import sys

# import Protocol and self return type type hints depending on Python version
PYTHON_VERSION = float("{:d}.{:d}".format(*sys.version_info[:2]))
if PYTHON_VERSION < 3.8:
    from typing_extensions import Protocol  # type: ignore
else:  # Python 3.8+
    from typing import Protocol  # type: ignore

from typing import Any, Dict

import numpy as np

# from typing import Protocol as Protocol  # type: ignore


class Predictor(Protocol):
    def predict(self, X: np.ndarray) -> np.ndarray:
        ...


class Model(Predictor, Protocol):
    def fit(self, X: np.ndarray, y: np.ndarray) -> Model:
        ...


class LinearModel(Model, Protocol):
    coef_: np.ndarray
    intercept_: float

    def fit(self, X: np.ndarray, y: np.ndarray) -> LinearModel:
        ...


class ModelCallback(Protocol):
    def __call__(self, **kwargs: Dict[str, Any]) -> Model:
        ...


class WeightsCallback(Protocol):
    def __call__(self, yt: np.ndarray, yp: np.ndarray) -> np.ndarray:
        ...



"""
Unit tests for SimplePLS weak learner class
"""

# author: Benjamin Cross
# email: btcross26@yahoo.com
# created: 2019-12-20


from unittest.mock import MagicMock

import numpy as np
import pytest

from genestboost.weak_learners import SimplePLS

from .simple_pls_data import X, y


@pytest.mark.parametrize("threshold", [None, 0.1, 0.5, 1.0])
def test_fit_max_vars_1(threshold):
    # GIVEN a SimplePLS instance with max_vars=1 and arbitrary filter_threshold
    model = SimplePLS(max_vars=1, filter_threshold=None)

    # WHEN the model is fit using provided dummy X, y data
    model.fit(X, y)

    # THEN only column 7 should have a coefficient (most correlated)
    assert model.coef_[7] != 0

    # THEN all other coefficients should be zero
    assert np.all(model.coef_[:7] == 0.0)

    # THEN the actual fitted column 7 coefficient and intercpet should be equal to the
    # least squares with one-var estimate using column 7
    Xs = np.hstack([np.ones(X.shape[0]).reshape(-1, 1), X[:, [7]]])
    ls_coef = np.linalg.lstsq(Xs, y, rcond=None)[0]
    np.testing.assert_allclose(
        np.array([model.intercept_, model.coef_[7]]), ls_coef, atol=0.0, rtol=1e-6
    )


@pytest.mark.parametrize("num_vars", [1, 2, 3, 4, 5, 6, 7, 8])
def test_fit_multiple_vars(num_vars):
    # GIVEN a SimplePLS instance with max_vars=num_vars and no filter_threshold
    model = SimplePLS(max_vars=num_vars, filter_threshold=None)

    # WHEN the model is fit using provided dummy X, y data
    model.fit(X, y)

    # THEN the number of non-zero coefs should equal num_vars
    assert (model.coef_ != 0.0).sum() == num_vars


def test_predict_method():
    # GIVEN a mocked SimplePLS object with predict method and known coefficients and
    # intercept attributes
    obj = MagicMock()
    obj.coef_ = np.array([0.0, 1.0, 2.0, 0.0, 3.0, -4.0, 0.0, -5.0])
    obj.intercept_ = 2.0
    obj.predict = SimplePLS.predict.__get__(obj)

    # WHEN the predict method is called on the mocked object with dummy data X
    calculated_values = obj.predict(X)

    # THEN the calculated_values should equal the values expected from a linear model
    expected_values = (
        2.0
        + 0.0 * X[:, 0]
        + 1.0 * X[:, 1]
        + 2.0 * X[:, 2]
        + 0.0 * X[:, 3]
        + 3.0 * X[:, 4]
        - 4.0 * X[:, 5]
        + 0.0 * X[:, 6]
        - 5.0 * X[:, 7]
    )
    np.testing.assert_allclose(calculated_values, expected_values, atol=0.0, rtol=1e-8)


@pytest.mark.parametrize("num_vars", [1, 2, 3])
def test_mask_coefs_max_vars(num_vars):
    # GIVEN a SimplePLS object fit on the X, y dummy data and num_vars equal to 1, 2,
    # or 3 (with no filter_threshold)
    model = SimplePLS(max_vars=num_vars, filter_threshold=None)

    # WHEN the model is fit on X, y
    model.fit(X, y)

    # THEN a 1-var model should only have non-zero coefficient column 7
    if num_vars == 1:
        assert model.coef_[7] != 0.0
        assert np.all(model.coef_[:7] == 0.0)

    # THEN a 2-var model should only have non-zero coefficient columns 7 and 0
    if num_vars == 2:
        assert np.all(model.coef_[[7, 0]] != 0.0)
        assert np.all(model.coef_[[1, 2, 3, 4, 5, 6]] == 0.0)

    # THEN a 3-var model should only have non-zero coefficient columns 7, 0, and 3
    if num_vars == 3:
        assert np.all(model.coef_[[7, 0, 3]] != 0.0)
        assert np.all(model.coef_[[1, 2, 4, 5, 6]] == 0.0)


@pytest.mark.parametrize("threshold", [1.0, 0.95, 0.60, 0.0])
def test_mask_coefs_filter_threshold(threshold):
    # GIVEN a SimplePLS object fit on the X, y dummy data with num_vars unspecified
    # and varying filter thresholds
    model = SimplePLS(max_vars=None, filter_threshold=threshold)

    # WHEN the model is fit on X, y
    model.fit(X, y)

    # THEN with 1.0 threshold, only column 7 should be selected
    if threshold == 1.0:
        assert model.coef_[7] != 0.0
        assert np.all(model.coef_[:7] == 0.0)

    # THEN with 0.95 threshold, only 7 and 0 should be selected
    if threshold == 0.95:
        assert np.all(model.coef_[[7, 0]] != 0.0)
        assert np.all(model.coef_[[1, 2, 3, 4, 5, 6]] == 0.0)

    # THEN with 0.60 threshold, only 7, 0, and 3 should be selected
    if threshold == 0.60:
        assert np.all(model.coef_[[7, 0, 3]] != 0.0)
        assert np.all(model.coef_[[1, 2, 4, 5, 6]] == 0.0)

    # THEN with 0.0 threshold, all variables should be non-zero
    if threshold <= 0.0:
        assert np.all(model.coef_ != 0.0)


def test_init_args_none():
    # GIVEN a SimplePLS object with None for both init args
    model = SimplePLS(max_vars=None, filter_threshold=None)

    # WHEN the model is fit on dummy data X, y
    model.fit(X, y)

    # THEN all variables should be used and non-zero
    assert np.all(model.coef_ != 0.0)



"""ModelDataSets class for splitting data into training and validation sets."""

# author: Benjamin Cross
# email: btcross26@yahoo.com
# created: 2019-12-26


from typing import Iterator, Optional

import numpy as np


class ModelDataSets:
    """ModelDataSets class for abstracting data set implementation from BoostedModel."""

    def __init__(
        self,
        X: np.ndarray,
        yt: np.ndarray,
        weights: Optional[np.ndarray] = None,
        validation_fraction: float = 0.0,
        validation_stratify: bool = False,
        random_state: Optional[int] = None,
    ):
        """
        Class initializer.

        Parameters
        ----------
        X: numpy.ndarray, shape (n_samples, n_features)
            Feature matrix of type float.

        yt: numpy.ndarray, shape (n_samples,)
            Target vector.

        weights: numpy.ndarray (optional, default=None), shape (n_samples,)
            Sample weights to be used in the fitting process.

        validation_fraction: float (optional, default=0.0)
            Fraction of dataset to use as validation set for early stopping.

        validation_stratify: bool (default=False)
            If True, stratify the validation sample and the training sample using the
            model target. This only makes sense for classification problems.

        random_state: int (optional, default=None)
            Set the random state of the instance so that the data set split can be
            reproduced.
        """
        # initialize attributes from init args
        self.X = X
        self.yt = yt
        self.weights = np.ones(yt.shape[0]) if weights is None else weights
        self.validation_fraction = validation_fraction
        self.validation_stratify = validation_stratify
        self._rng = np.random.RandomState(random_state)

        # public vars to be created during class usage
        self.X_train: np.ndarray
        self.yt_train: np.ndarray
        self.weights_train: np.ndarray
        self.X_val: np.ndarray  # if validation_fraction > 0.0
        self.yt_val: np.ndarray  # if validation_fraction > 0.0
        self.weights_val: np.ndarray  # if validation_fraction > 0.0

        # private vars to be created during class usage
        self._tindex: np.ndarray
        self._vindex: np.ndarray  # if validation_fraction > 0.0

        # split data sets as necessary
        self._create_index(validation_fraction, validation_stratify)
        self._create_data_sets()

    def has_validation_set(self) -> bool:
        """Return True if the validation fraction is greater than zero."""
        return hasattr(self, "_vindex")

    def _create_data_sets(self) -> None:
        """Create the train/test datasets (private)."""
        self.X_train = self.X[self._tindex]
        self.yt_train = self.yt[self._tindex]
        self.weights_train = self.weights[self._tindex]

        if self.has_validation_set():
            self.X_val = self.X[self._vindex]
            self.yt_val = self.yt[self._vindex]
            self.weights_val = self.weights[self._vindex]

    def _create_index(
        self, validation_fraction: float, validation_stratify: bool
    ) -> None:
        """Get the sampled train/test indices (private method)."""
        # initialize training index
        self._tindex = np.arange(self.yt.shape[0]).astype(int)

        # create validation index if specified
        if validation_fraction > 0.0:
            if validation_stratify:
                self._vindex = np.hstack(
                    [
                        array
                        for array in self._stratify_groups_generator(
                            validation_fraction
                        )
                    ]
                )
            else:
                n = self.yt.shape[0]
                self._vindex = np.random.choice(
                    n, int(validation_fraction * n), replace=False
                )
            self._vindex.sort()
            self._tindex = np.setdiff1d(self._tindex, self._vindex, assume_unique=True)

            # convert to int
            self._vindex = self._vindex.astype(int)
            self._tindex = self._tindex.astype(int)

    def _stratify_groups_generator(
        self, validation_fraction: float
    ) -> Iterator[np.array]:
        """Stratify the dataset for classification problems (private method)."""
        full_index = np.arange(self.yt.shape[0])
        groups, group_index = np.unique(self.yt, return_inverse=True)
        for group in groups:
            mask = group_index == group
            n = mask.sum()
            g_index = self._rng.choice(
                full_index[mask], int(validation_fraction * n), replace=False
            )
            yield g_index



"""Link functions related to power function."""

# author: Benjamin Cross
# email: btcross26@yahoo.com
# created: 2019-08-26


from typing import Union

import numpy as np

from .base_class import BaseLink


class PowerLink(BaseLink):
    """Implementation of power link functions."""

    def __init__(self, power: Union[float, int], summand: float = 0.0):
        """
        Initialize a power link instance - i.e., link = (y + summand) ** power.

        Extends BaseLink class intializer.

        Parameters
        ----------
        power: float or int
            Exponent of the link function.

        summand: float or int
            Summand of the link function.
        """
        super().__init__()
        self.power_ = power
        self.summand_ = summand

    def _link(self, y: np.ndarray) -> np.ndarray:
        """
        Get the link, eta, as a function of y.

        Overrides BaseLink._link.
        """
        return (y + self.summand_) ** self.power_

    def _inverse_link(self, eta: np.ndarray) -> np.ndarray:
        """
        Get the target, y, as a function of the link, `eta`.

        Overrides BaseLink._inverse_link.
        """
        return eta ** (1.0 / self.power_) - self.summand_

    def dydeta(self, y: np.ndarray) -> np.ndarray:
        """
        Get the derivative of `y` with respect to the link as a function of y.

        Overrides BaseLink.dydeta.
        """
        return (1.0 / self.power_) * (y + self.summand_) ** (1.0 - self.power_)

    def d2ydeta2(self, y: np.ndarray) -> np.ndarray:
        """
        Get the second derivative of `y` with respect to the link as a function of y.

        Overrides BaseLink.d2ydeta2.
        """
        return self.dydeta(y) * (1.0 / self.power_ - 1.0) / self._link(y)


class SqrtLink(PowerLink):
    """Square root link function implementation."""

    def __init__(self) -> None:
        """
        Class initializer.

        Extends PowerLink class intializer by specifying power=0.5 and summand=0.0.
        """
        super().__init__(power=0.5)


class CubeRootLink(PowerLink):
    """Cube root link function implementation."""

    def __init__(self) -> None:
        """
        Class initializer.

        Extends PowerLink class initializer by specifying power=1/3 and summand=0.0.
        """
        super().__init__(power=(1.0 / 3.0))


class ReciprocalLink(PowerLink):
    """Inverse link function implementation."""

    def __init__(self, summand: float = 0.0):
        """
        Class initializer.

        Extends PowerLink class initializer by specifying power=-1 and summand=0.0.
        """
        super().__init__(power=-1.0, summand=summand)



"""
Unit tests for BoostedModel.InitialModel class
"""

# author: Benjamin Cross
# email: btcross26@yahoo.com
# created: 2020-04-06


import logging

import numpy as np
import pytest

from genestboost import BoostedModel
from genestboost.link_functions import IdentityLink

from .weak_learners.simple_pls_data import X, y

LOGGER = logging.getLogger(__name__)


# test init types
@pytest.mark.parametrize(
    "init_type", ["zero", "residuals", "mean"], ids=["zero", "residuals", "mean"],
)
def test_init_types(init_type):
    """
    Quick obvious tests on initial values in InitialModel class
    """
    # GIVEN a BoostedModel instance with internal link, loss, and model_callbacks
    model = BoostedModel.InitialModel(link=IdentityLink(), init_type=init_type)

    # WHEN fitting the model
    model.fit(X, y)

    # THEN appropriate attributes should be stored based on init_type
    # when weights are not specified
    if init_type in ["zero", "residuals"]:
        assert model._value == 0.0
    elif init_type == "mean":
        assert model._value == model._link(np.mean(y))


# test initial model exception
@pytest.mark.parametrize("init_type", [None, "other", "error", 5])
def test_initial_model_exception(init_type):
    """
    Test exception for BoostedModel.InitialModel when init_type is mis-specified
    """
    # GIVEN a BoostedModel instance with internal link, loss, and model_callbacks
    # and errant specification of init_type
    model = BoostedModel.InitialModel(link=IdentityLink(), init_type=init_type)

    # WHEN fitting the model
    # THEN an exception should be thrown
    with pytest.raises(AttributeError) as excinfo:
        model.fit(X, y)
    assert excinfo.value.args[0] == "init arg:<init_type> is mis-specified"


# test predict method of InitialModel
@pytest.mark.parametrize("init_type", [None, "other", "error", 5])
def test_initial_model_exception(init_type):
    """
    Test exception for BoostedModel.InitialModel when init_type is mis-specified
    """
    # GIVEN a BoostedModel instance with internal link, loss, and model_callbacks
    # and errant specification of init_type
    model = BoostedModel.InitialModel(link=IdentityLink(), init_type=init_type)

    # WHEN fitting the model
    # THEN an exception should be thrown
    with pytest.raises(AttributeError) as excinfo:
        model.fit(X, y)
    assert excinfo.value.args[0] == "init arg:<init_type> is mis-specified"



"""Complementary log-log link function implementation."""

# author: Benjamin Cross
# email: btcross26@yahoo.com
# created: 2019-08-26


import numpy as np

from .base_class import BaseLink


class CLogLogLink(BaseLink):
    """Complementary log-log link function implementation."""

    def __init__(self, eps: float = 1e-10):
        """
        Class initializer.

        Extends the BaseLink class initializer.

        Parameter
        ---------
        eps: float (default = 1e-10)
            A small constant float to prevent log from returning negative infinity.
        """
        super().__init__()
        self._eps = eps

    def _link(self, y: np.ndarray) -> np.ndarray:
        """
        Get the link, eta, as a function of y.

        Overrides BaseLink._link.
        """
        return np.log(-np.log(1.0 - y + self._eps))

    def _inverse_link(self, eta: np.ndarray) -> np.ndarray:
        """
        Get the target, y, as a function of the link, `eta`.

        Overrides BaseLink._inverse_link.
        """
        return 1.0 - np.exp(-np.exp(eta))

    def dydeta(self, y: np.ndarray) -> np.ndarray:
        """
        Get the derivative of `y` with respect to the link as a function of y.

        Overrides BaseLink.dydeta.
        """
        return -(1.0 - y) * np.log(1.0 - y + self._eps)

    def d2ydeta2(self, y: np.ndarray) -> np.ndarray:
        """
        Get the second derivative of `y` with respect to the link as a function of y.

        Overrides BaseLink.d2ydeta2.
        """
        return self.dydeta(y) * (1.0 + np.log(1.0 - y + self._eps))



"""
Tests for functions in genestboost.utils gradients.py src file
"""

# author: Benjamin Cross
# email: btcross26@yahoo.com
# created: 2019-12-16


import numpy as np

from genestboost.utils import d1_central_difference, d2_central_difference


# tests for central difference utils functions
class TestCentralDifferences:
    @staticmethod
    def f(y):
        """
        Arbitrary polynomial function
        """
        return 3.0 * y ** 3 - 4.0 * y ** 2 + 7.0 * y - 4.0

    @staticmethod
    def fp1(y):
        """
        First derivative of f
        """
        return 9.0 * y ** 2 - 8.0 * y + 7.0

    @staticmethod
    def fp2(y):
        """
        Second derivative of f
        """
        return 18.0 * y - 8.0

    def test_d1_central_difference(self):
        """
        Test first-order first derivative utils function
        """
        # GIVEN a simple polynomial function, derivatives, and evaluation locations
        y = np.linspace(2, 12, 101)

        # WHEN the first order central difference is used to approximate derivatives
        expected_values_d1 = self.fp1(y)
        expected_values_d2 = self.fp2(y)
        calculated_values_d1 = d1_central_difference(self.f, y, 1e-8)
        calculated_values_d2 = d1_central_difference(self.fp1, y, 1e-8)

        # THEN calculate values should be reasonable close to exact values
        np.testing.assert_allclose(
            calculated_values_d1, expected_values_d1, atol=0.0, rtol=1e-3
        )
        np.testing.assert_allclose(
            calculated_values_d2, expected_values_d2, atol=0.0, rtol=1e-3
        )

    def test_d2_central_difference(self):
        """
        Test first-order second derivative utils function
        """
        # GIVEN a simple polynomial function, derivatives, and evaluation locations
        y = np.linspace(2, 12, 101)

        # WHEN the first order central difference is used to approximate 2nd derivatives
        expected_values_d2 = self.fp2(y)
        calculated_values_d2 = d2_central_difference(self.f, y, 1e-6)

        # THEN calculate values should be reasonable close to exact values
        np.testing.assert_allclose(
            calculated_values_d2, expected_values_d2, atol=0.0, rtol=1e-2
        )



"""
Unit tests for link function implementations
"""

# author: Benjamin Cross
# email: btcross26@yahoo.com
# created: 2019-12-16

import logging

import numpy as np
import pytest

from genestboost.utils import d1_central_difference, d2_central_difference

from .link_test_list import LINK_TESTS

LOGGER = logging.getLogger(__name__)


# constants
TEST_NAMES = list(map(lambda x: x[0], LINK_TESTS))


# generic class to test link functions
@pytest.mark.parametrize(
    "test_name,link,link_test_values,d_test_values,tol", LINK_TESTS, ids=TEST_NAMES
)
class TestLinkFunction:
    """
    Generic unit tests for link functions
    """

    def test_link_function(self, test_name, link, link_test_values, d_test_values, tol):
        """
        Test link calculations
        """
        # GIVEN some true (y, nu) pairs of link function values
        y = link_test_values[:, 0]
        eta = link_test_values[:, 1]

        # WHEN the link function is called on x
        calculated_values = link(y)

        # THEN the correct values of nu should be returned
        np.testing.assert_allclose(calculated_values, eta)

    def test_inverse_link_function(
        self, test_name, link, link_test_values, d_test_values, tol
    ):
        """
        Test inverse link calculations
        """
        # GIVEN some true (y, nu) pairs of link function values
        y = link_test_values[:, 0]
        eta = link_test_values[:, 1]

        # WHEN the inverse link function is called on nu
        calculated_values = link(eta, inverse=True)

        # THEN the correct values of x should be returned
        np.testing.assert_allclose(calculated_values, y)

    def test_d1_link(self, test_name, link, link_test_values, d_test_values, tol):
        """
        Test link first derivative calculations
        """
        # GIVEN some y locations to calculate link derivatives
        y = d_test_values
        eta = link(d_test_values)

        # WHEN the first derivative is approximated using central differences
        def func(eta):
            return link(eta, inverse=True)

        calculated_values = d1_central_difference(func, eta, h=1e-8)

        # THEN the calculated values of the derivative should be close to the
        # true implemented computed values
        expected_values = link.dydeta(y)
        np.testing.assert_allclose(
            calculated_values, expected_values, atol=tol[0], rtol=tol[1]
        )

    def test_d2_link(self, test_name, link, link_test_values, d_test_values, tol):
        """
        Test link second derivative calculations
        """
        # GIVEN some y locations to calculate link derivatives
        y = d_test_values
        eta = link(d_test_values)

        # WHEN the second derivative is approximated using central differences
        def func(eta):
            return link(eta, inverse=True)

        calculated_values = d2_central_difference(func, eta, h=1e-6)

        # THEN the calculated values of the derivative should be close to the
        # true implemented computed values
        expected_values = link.d2ydeta2(y)
        np.testing.assert_allclose(
            calculated_values, expected_values, atol=tol[0], rtol=tol[1]
        )



"""
Unit tests for beta loss class implementations. Checks that the beta loss
implementation matches quasi-deviance with beta callback and that the leaky
beta loss matches intent.
"""

# author: Benjamin Cross
# email: btcross26@yahoo.com
# created: 2019-12-19


import numpy as np
import pytest

from genestboost.loss_functions import BetaLoss, LeakyBetaLoss, QuasiLogLoss

# module pytest fixtures

# fixture for beta loss tests
@pytest.fixture(
    scope="function",
    params=[(2.0, 2.0), (4.0, 2.0), (3.0, 9.0)],
    ids=["alpha=2,beta=2", "alpha=4,beta=2", "alpha=3,beta=9"],
)
def beta_loss_objects(request):
    # setup
    alpha, beta = request.param
    beta_callback = BetaLoss.beta_callback(alpha, beta)
    ql_loss = QuasiLogLoss(beta_callback, d0_n=5000)
    beta_loss = BetaLoss(alpha, beta)

    # yield objects
    yield ql_loss, beta_loss

    # teardown
    del beta_callback, ql_loss, beta_loss


# fixture for leaky beta loss tests
@pytest.fixture(
    scope="function",
    params=[0.01, 0.1, 0.5, 1.0],
    ids=["gamma=0.01", "gamma=0.1", "gamma=0.5", "gamma=1.0"],
)
def leaky_beta_loss_objects(request):
    # setup
    alpha, beta = 3.0, 9.0
    gamma = request.param
    leaky_beta_loss = LeakyBetaLoss(alpha, beta, gamma)
    beta_loss = BetaLoss(alpha, beta)

    # yield objects
    yield leaky_beta_loss, beta_loss

    # teardown
    del leaky_beta_loss, beta_loss


# tests for BetaLoss class
def test_beta_loss(beta_loss_objects):
    # GIVEN a QuasiLogLoss instance with beta denominator callback and
    # BetaLoss instance with equivalent alpha/beta parameters
    ql_loss, beta_loss = beta_loss_objects

    # WHEN loss values are calculated at various points
    yt = np.array([1.0, 1.0, 1.0, 0.0, 0.0, 0.0])
    yp = np.array([0.1, 0.5, 0.9, 0.1, 0.5, 0.9])
    beta_calc = beta_loss(yt, yp)
    ql_calc = ql_loss(yt, yp)

    # THEN the two different instances should yield approximately equal values
    np.testing.assert_allclose(beta_calc, ql_calc, atol=1e-3, rtol=1e-4)

    # WHEN d1 values are calculated at various points
    beta_calc_d1 = beta_loss.dldyp(yt, yp)
    ql_calc_d1 = ql_loss.dldyp(yt, yp)

    # THEN the two different instances should yield approximately equal values
    np.testing.assert_allclose(beta_calc_d1, ql_calc_d1, atol=1e-4, rtol=1e-5)

    # WHEN d2 values are calculated at various points
    beta_calc_d2 = beta_loss.d2ldyp2(yt, yp)
    ql_calc_d2 = ql_loss.d2ldyp2(yt, yp)

    # THEN the two different instances should yield approximately equal values
    np.testing.assert_allclose(beta_calc_d2, ql_calc_d2, atol=1e-3, rtol=1e-4)


# tests for LeakyBetaLoss class
def test_leaky_beta_loss(leaky_beta_loss_objects):
    # GIVEN a LeakyBetaLoss instance with sister BetaLoss instance
    leaky_loss, beta_loss = leaky_beta_loss_objects

    # WHEN loss values are calculated at various points
    yt = np.array([1.0, 1.0, 1.0, 0.0, 0.0, 0.0])  # mid-values
    yp = np.array([0.3, 0.5, 0.9, 0.1, 0.15, 0.2])  # mid-values
    leaky_calc = leaky_loss(yt, yp)
    beta_calc = beta_loss(yt, yp)

    # THEN the two different instances should yield equal values inside of the
    # ratio points
    np.testing.assert_allclose(leaky_calc, beta_calc, atol=0.0, rtol=1e-8)

    # WHEN d1 values are calculated at various points inside the ratio points
    beta_calc_d1a = beta_loss.dldyp(yt, yp)
    leaky_calc_d1a = leaky_loss.dldyp(yt, yp)

    # THEN the two different instances should yield equal values
    np.testing.assert_allclose(leaky_calc_d1a, beta_calc_d1a, atol=0.0, rtol=1e-8)

    # WHEN d1 values are calculated at various points outside the ratio points
    yt_shelf = np.array([1.0, 0.0, 0.0])  # shelf-values
    yp_shelf = np.array([0.001, 0.9, 0.99])  # shelf-values
    beta_calc_d1b = leaky_loss.gamma * beta_loss.dldyp(yt_shelf, 0.25)
    leaky_calc_d1b = leaky_loss.dldyp(yt_shelf, yp_shelf)

    # THEN the two different instances should yield equal values
    np.testing.assert_allclose(leaky_calc_d1b, beta_calc_d1b, atol=0.0, rtol=1e-8)

    # WHEN d2 values are calculated at various points inside ratio points
    beta_calc_d2a = beta_loss.d2ldyp2(yt, yp)
    leaky_calc_d2a = leaky_loss.d2ldyp2(yt, yp)

    # THEN the two sets should be equivalent
    np.testing.assert_allclose(leaky_calc_d2a, beta_calc_d2a, atol=0.0, rtol=1e-8)

    # WHEN d2 values are calculated at various points outside ratio points
    leaky_calc_d2b = leaky_loss.d2ldyp2(yt_shelf, yp_shelf)

    # THEN the values should equal ZERO
    np.testing.assert_allclose(leaky_calc_d2b, 0.0, atol=0.0, rtol=1e-8)


# test for LeakyBetaLoss class transition points
def test_leaky_beta_loss_transition(leaky_beta_loss_objects):
    # GIVEN a LeakyBetaLoss instance
    leaky_loss, _ = leaky_beta_loss_objects

    # WHEN loss values are calculated just to the left and just to the right of the
    # transition points (and first derivative)
    eps = 1e-8

    # yp/yt values
    yp_left = np.array([leaky_loss.xL - eps, leaky_loss.xR - eps])
    yp_right = np.array([leaky_loss.xL + eps, leaky_loss.xR + eps])
    yt = np.array([0.0, 1.0])

    # loss
    loss_left = leaky_loss(yt, yp_left)
    loss_right = leaky_loss(yt, yp_right)

    # dldyp
    dl_left = leaky_loss.dldyp(yt, yp_left)
    dl_right = leaky_loss.dldyp(yt, yp_right)

    # THEN the two calculations should be very close to one another for both
    # loss and derivatives
    np.testing.assert_allclose(loss_left, loss_right, atol=1e-6, rtol=0.0)
    np.testing.assert_allclose(dl_left, dl_right, atol=1e-6, rtol=0.0)


# test for LeakyBetaLoss class end points
def test_leaky_beta_loss_transition(leaky_beta_loss_objects):
    # GIVEN a LeakyBetaLoss instance
    leaky_loss, _ = leaky_beta_loss_objects

    # WHEN loss values are calculated close to the end points and at the
    # transition points
    eps = 1e-8

    # yp/yt values
    yp_transition = np.array([leaky_loss.xL, leaky_loss.xR])
    yp_end = np.array([1.0 - eps, eps])
    yt = np.array([0.0, 1.0])

    # loss
    loss_transition = leaky_loss(yt, yp_transition)
    loss_end = leaky_loss(yt, yp_end)

    # THEN the end point calculations should be greater than or equal to the transition
    # point calculations
    assert np.all(loss_end >= loss_transition)



"""General boosting model class implementation for any regressor."""

# author: Benjamin Cross
# email: btcross26@yahoo.com
# created: 2019-08-26


import logging
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np

from .link_functions import BaseLink
from .loss_functions import BaseLoss
from .model_data_sets import ModelDataSets
from .type_hints import Model, ModelCallback, WeightsCallback

LOGGER = logging.getLogger(__name__)


# BoostedModel implementation
class BoostedModel:
    """General boosting model class implementation for any regression model."""

    def __init__(
        self,
        link: BaseLink,
        loss: BaseLoss,
        model_callback: ModelCallback,
        model_callback_kwargs: Optional[Dict[str, Any]] = None,
        weights: Union[str, WeightsCallback] = "none",
        alpha: float = 1.0,
        step_type: str = "decaying",
        step_decay_factor: float = 0.48,
        init_type: str = "mean",
        random_state: Optional[int] = None,
        validation_fraction: float = 0.0,
        validation_stratify: bool = False,
        validation_iter_stop: int = 10,
        tol: Optional[float] = None,
    ):
        """
        Class initializer.

        Parameters
        ----------
        link: BaseLink
            Link function to use in boosting iterations.

        loss: BaseLoss
            Loss function to use in boosting iterations.

        model_callback: Callable
            A callable that returns a model object that implements fit and predict methods.

        model_callback_kwargs: dict, optional (default=None)
            A dictionary of keyword arguments to pass to `model_callback`.

        weights: Union[str, WeightsCallback]: str or Callable
            A string specificying the type of weights (one of "none" or "newton"), or a
            callable of the form 'lambda yt, yp: weights`, where yt and yp are the
            actual target and predicted target arrays. These weights are multiplied
            element-wise by the model gradients to produce the pseudo-residuals that are
            to be predicted at each model iteration.

        alpha: float = 1.0
            A parameter representing the intial trial learning rate. The learning rate that
            actually gets used at each iteration is dependent upon the value of `step_type`.

        step_type: str (default="decaying")
            One of "decaying", "constant", or "best". For "decaying", the initial model
            iteration will start with `alpha` as the learning parameter. If `alpha` times
            `step_decay_factor` results in a greater reduction in loss than `alpha`, then
            use the former. This is repeated until performance does not improve, and the
            final chosen rate will serve as the 'initial' rate for the next boosting
            iteration. For "constant", `alpha` will be used at every boosting iteration.
            Using "step_type" best implements the same process as "decaying", except the
            next boosting iteration will reset the learning rate back to `alpha` as the
            initial trial learning rate.

        step_decay_factor: float (default=0.48)
            The decaying factor to use with `step_type` "decaying" or "best".

        init_type: str (default="mean")
            The type of intial prediction to use. If "mean", then the initial link
            prediction (prior to boosting iterations) will be taken as the link of the mean
            of the non-transformed target. If "residuals" or "zero", the initial link
            prediction will be set to 0.0.

        random_state: int, optional (default=None)
            Random seed to be used for reproducability when random methods are used
            internally.

        validation_fraction: float (default=0.0)
            If 0.0, then no validation set will be used and training will be performed
            using the full training set. Otherwise, the fraction of observations to use
            as a holdout set for early stopping.

        validation_stratify: bool (default=False)
            If true, than stratify the validation holdout set based on the target. Useful
            when the target is binary.

        validation_iter_stop: int (default=10)
            Number of iterations to use for early stopping on the validation set. If the
            holdout loss is greater at the current iteration than `validation_iter_stop`
            iterations prior, then stop model fitting.

        tol: float, optional (default=None)
            Early stopping criteria based on training loss. If training loss fails to
            improve by at least `tol`, then stop training. If None, then training loss
            criteria is not checked to determine early stopping.
        """
        # set state based on initializer arguments
        self._link = link
        self._loss = loss
        self.model_callback = model_callback
        self.model_callback_kwargs = (
            dict() if model_callback_kwargs is None else model_callback_kwargs
        )  # type: Dict[str, Any]
        self.weights = weights
        self.alpha = alpha
        self.step_type = step_type
        self.step_decay_factor = step_decay_factor
        self.init_type = init_type
        self.random_state = random_state
        self.validation_fraction = validation_fraction
        self.validation_stratify = validation_stratify
        self.validation_iter_stop = validation_iter_stop
        self.tol = tol

        # additional vars used during the fitting process
        self._beta: float = alpha
        self._msi = -1 if init_type in ["offset", "residuals"] else None
        self._loss_list: List[Tuple[float, float]] = list()
        self._model_list: List[Tuple[Model, float]] = list()
        self._is_fit: bool = False
        self._model_init: BoostedModel.InitialModel

    def __bool__(self) -> bool:
        """Get the model fit status.

        Returns
        =======
        bool
            True if the model has been fit, false otherwise.
        """
        return self._is_fit

    def boost(
        self,
        X: np.ndarray,
        yt: np.ndarray,
        yp: np.ndarray,
        eta_p: np.ndarray,
        model_callback: ModelCallback,
        model_callback_kwargs: Dict[str, Any],
        weights: Optional[np.ndarray] = None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Boost by one model iteration.

        Creates a model using the `model_callback` callable with `model_callback_kwargs`,
        then fits this model to the pseudo-residuals. The learning rate is determined
        according to the chosen method, and current predictions are updated and returned
        such that stopping criteria can be evaluated and boosting continued. The
        fitted model resulting from the iteration is appended to the underlying model
        ensemble.

        Parameters
        ----------
        X: np.ndarray
            The model matrix. If `init_type` was set as "residuals", then the model
            scores for which to calculate residuals should form the last column of the
            input matrix.

        yt: np.ndarray
            Observed target values.

        yp: np.ndarray
            Predicted target values.

        eta_p: np.ndarray
            The current link predictions corresponding to the model iteration. This can
            be found by transforming `yp`, but passing this as an argument avoids
            duplicate computations and improves performance.

        model_callback: Callable
            A callable that returns a model object that implements fit and predict methods.

        model_callback_kwargs: dict, optional (default=None)
            A dictionary of keyword arguments to pass to `model_callback`.

        weights: np.ndarray, optional (default=None)
            Sample weights (by observation) to use for fitting. Should be positive.
            Observations with higher weights will affect the model fit more. If 'None',
            then all weights will be equal (1.0).

        Returns
        -------
        yp_next, eta_p_next: tuple(np.ndarray, np.ndarray)
            A tuple of the updated target predictions and target prediction links.
        """
        model_ = model_callback(**model_callback_kwargs)
        weights = 1.0 if weights is None else weights
        p_residuals = self.compute_p_residuals(yt, yp) * weights
        model_ = model_.fit(X[:, : self._msi], p_residuals)
        preds = model_.predict(X[:, : self._msi])
        learning_rate = self._compute_beta(yt, eta_p, preds)
        eta_p_next = eta_p + learning_rate * preds
        yp_next = self._link(eta_p_next, inverse=True)
        self._model_list.append((model_, learning_rate))
        return yp_next, eta_p_next

    def compute_link(self, yp: np.ndarray, inverse: bool = False) -> np.ndarray:
        """Compute the element-wise link function or inverse link function.

        Parameters
        ----------
        yt: np.ndarray
            Observed target values.

        yp: np.ndarray
            Predicted target values.

        Returns
        -------
        np.ndarray
        """
        return self._link(yp, inverse)

    def compute_loss(self, yt: np.ndarray, yp: np.ndarray) -> np.ndarray:
        """Compute element-wise loss.

        Parameters
        ----------
        yt: np.ndarray
            Observed target values.

        yp: np.ndarray
            Predicted target values.

        Returns
        -------
        np.ndarray
        """
        return self._loss(yt, yp)

    def compute_gradients(self, yt: np.ndarray, yp: np.ndarray) -> np.ndarray:
        """Compute element-wise gradients.

        Parameters
        ----------
        yt: np.ndarray
            Observed target values.

        yp: np.ndarray
            Predicted target values.

        Returns
        -------
        np.ndarray
        """
        return self._loss.dldyp(yt, yp) * self._link.dydeta(yp)

    def compute_newton_weights(self, yt: np.ndarray, yp: np.ndarray) -> np.ndarray:
        """
        Compute newton weights.

        This is a mixture of loss function derivatives and link function derivatives
        by application of chain rule. Using newton weights requires that the second
        derivatives of the loss and link function be defined. This method uses the
        computed second derivatives as-is - there are no adjustments to prevent
        the effects of ill-conditioning (very small second derivatives) or non-positive
        definiteness (negative second derivatives) on computed pseudo residuals.

        Parameters
        ----------
        yt: np.ndarray
            Observed target values.

        yp: np.ndarray
            Predicted target values.

        Returns
        -------
        np.ndarray
            The element-wise reciprocal of the second-derivative of the loss function
            with respect to the link function.
        """
        term_1 = self._loss.d2ldyp2(yt, yp) * self._link.dydeta(yp) ** 2
        term_2 = self._loss.dldyp(yt, yp) * self._link.d2ydeta2(yp)
        denominator = term_1 + term_2
        return 1.0 / denominator

    def compute_weights(self, yt: np.ndarray, yp: np.ndarray) -> np.ndarray:
        """
        Compute model weights that will be multiplied by observation gradients.

        The final result is the pseudo-residuals to be fit at the next boosting
        iteration. This essentially serves as a case/switch statement to redirect to
        the underlying calculation method.

        Parameters
        ----------
        yt: np.ndarray
            Observed target values.

        yp: np.ndarray
            Predicted target values.
        """
        if self.weights == "none":
            return 1.0

        if self.weights == "newton":
            return self.compute_newton_weights(yt, yp)

        if callable(self.weights):
            return self.weights(yt, yp)

        raise AttributeError(
            "attribute:<weights> should be 'none', 'newton', or a callable"
        )

    def compute_p_residuals(self, yt: np.ndarray, yp: np.ndarray) -> np.ndarray:
        """
        Calculate pseudo-residuals.

        The psuedo-residuals are taken as the observation gradients times weights that
        are computed as per the selected weighting scheme ("none", "newton", callable).

        Parameters
        ----------
        yt: np.ndarray
            Observed target values.

        yp: np.ndarray
            Predicted target values.

        Returns
        -------
        np.ndarray
        """
        numerator = -self.compute_gradients(yt, yp)
        denominator = self.compute_weights(yt, yp)
        return numerator / denominator

    def decision_function(
        self, X: np.ndarray, model_index: Optional[int] = None
    ) -> np.ndarray:
        """
        Get the link of computed model predictions.

        Parameters
        ----------
        X: np.ndarray
            The model matrix. If `init_type` was set as "residuals", then the model
            scores for which to calculate residuals should form the last column of the
            input matrix.

        model_index: int, optional (default=None)
            If None, then return the full model prediction as the sum of all models in
            the ensemble plus the initial model prediction. If an int, then return only
            the predictions from models up to `model_index` (i.e., [:model_index]).

        Returns
        -------
        np.ndarray
            The link of the computed model predictions.
        """
        eta_p = self._model_init.predict(X)
        for model, lr in self._model_list[:model_index]:
            eta_p += lr * model.predict(X[:, : self._msi])
        return eta_p

    def decision_function_single(
        self, X: np.ndarray, model_index: int = -1, apply_learning_rate: bool = True,
    ) -> np.ndarray:
        """
        Compute the link for a specific ensemble model by index.

        Parameters
        ----------
        X: np.ndarray
            The model matrix. If `init_type` was set as "residuals", then the model
            scores for which to calculate residuals should form the last column of the
            input matrix.

        model_index: int (default=-1)
            The model iteration for which to compute the decision function. By default,
            it is -1. This corresponds to the model from the most recent boosting
            iteration.

        apply_learning_rate: bool (default=True)
            If True, then the predictions from the selected model on `X` will be
            multiplied by the corresponding learning rate. Otherwise if False, the
            predictions of the selected model will be returned as if the learning rate
            was equal to 1.0.

        Returns
        -------
        np.ndarray
            The computed link values for the selected model index.
        """
        model, lr = self._model_list[model_index]
        lr = lr if apply_learning_rate else 1.0
        eta_p = lr * model.predict(X[:, : self._msi])
        return eta_p

    def fit(
        self,
        X: np.ndarray,
        yt: np.ndarray,
        iterations: int = 100,
        weights: Optional[np.ndarray] = None,
        min_iterations: Optional[int] = None,
    ) -> Model:
        """
        Fit the boosted model.

        Parameters
        ----------
        X: np.ndarray
            The model matrix. If `init_type` was set as "residuals", then the model
            scores for which to calculate residuals should form the last column of the
            input matrix.

        iterations: int (default=100)
            The maximum number of boosting iterations to perform.

        weights: np.ndarray, optional (default=None)
            Sample weights (by observation) to use for fitting. Should be positive.
            Observations with higher weights will affect the model fit more. If 'None',
            then all weights will be equal (1.0).

        min_iterations: int, optional (default=None)
            The minimum number of boosting iterations to perform. If None (the default),
            then there is no minimum.

        Returns
        -------
        self
        """
        # compute weights if null and create modeling data sets
        weights = np.ones_like(yt) if weights is None else weights
        model_data = ModelDataSets(
            X,
            yt,
            weights,
            self.validation_fraction,
            self.validation_stratify,
            self.random_state,
        )

        # initialize model and get initial predictions
        yp_train, eta_p_train = self.initialize_model(
            model_data.X_train, model_data.yt_train, model_data.weights_train
        )

        # get validation set predictions
        if model_data.has_validation_set():
            eta_p_val = self.decision_function(model_data.X_val)
            yp_val = self._link(eta_p_val, inverse=True)
        else:
            yp_val = None

        # calculate initial loss if model has not started
        if self.get_iterations() == 0:
            self._track_loss(yp_train, yp_val, model_data)

        # perform boosting iterations
        min_iterations = 0 if min_iterations is None else min_iterations
        for i in range(iterations):
            # perform boosting step
            yp_train, eta_p_train = self.boost(
                model_data.X_train,
                model_data.yt_train,
                yp_train,
                eta_p_train,
                self.model_callback,
                self.model_callback_kwargs,
                model_data.weights_train,
            )

            # track loss
            if model_data.has_validation_set():
                last_model, lr = self._model_list[-1]
                eta_p_val = eta_p_val + lr * last_model.predict(
                    model_data.X_val[:, : self._msi]
                )
                yp_val = self._link(eta_p_val, inverse=True)
            self._track_loss(yp_train, yp_val, model_data)

            # check stopping criteria
            if i + 1 > min_iterations and self._stop_model(model_data):
                break

        return self

    def get_model_links(self, X: np.ndarray) -> np.ndarray:
        """
        Compute a matrix of model links (without applying learning rates).

        Returns a matrix of model links, where each column index corresponds to the
        same index in the model ensemble.

        Parameters
        ----------
        X: np.ndarray [n_samples, n_features]
            The model matrix. If `init_type` was set as "residuals", then the model
            scores for which to calculate residuals should form the last column of the
            input matrix.

        Returns
        -------
        link_matrix: np.ndarray [n_samples, n_boosting_iterations]
        """
        model_links = np.zeros(
            (X.shape[0], len(self._model_list) + 1), dtype=np.float64
        )
        eta_p = self._model_init.predict(X)
        model_links[:, 0] = eta_p
        for i, (model, _) in enumerate(self._model_list):
            eta_p = model.predict(X[:, : self._msi])
            model_links[:, i + 1] = eta_p
        return model_links

    def get_loss_history(self) -> np.ndarray:
        """
        Get the loss history for the fitted model (training and validation loss).

        Returns
        -------
        np.ndarray
            A two-column array with with training and holdout loss in each column,
            respectively.
        """
        return np.array(self._loss_list).astype(np.float)

    def get_iterations(self) -> int:
        """Get the current number of model boosting iterations."""
        return len(self._model_list)

    def initialize_model(
        self, X: np.ndarray, yt: np.ndarray, weights: Optional[np.ndarray] = None
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Initialize the boosted model.

        This is called internally within the fit function. If manual boosting is being
        performed using the 'boost' method, then this method should be called before
        beginning the manual boosting procedure.

        Parameters
        ----------
        X: np.ndarray [n_samples, n_features]
            The model matrix. If `init_type` was set as "residuals", then the model
            scores for which to calculate residuals should form the last column of the
            input matrix.

        yt: np.ndarray
            Observed target values.

        weights: np.ndarray, optional (default=None)
            Sample weights (by observation) to use for fitting. Should be positive.
            Observations with higher weights will affect the model fit more. If 'None',
            then all weights will be equal (1.0).

        Returns
        -------
        yp, eta_p: tuple(np.ndarray, np.ndarray)
            The initial target predictions and link of target prediction arrays.
        """
        if not self._is_fit:
            # initialize model lists/attributes
            self._loss_list = list()
            self._model_list = list()
            self._model_init = self.InitialModel(self._link, self.init_type)
            self._model_init.fit(X, yt, weights)

            # set _is_fit to True
            self._is_fit = True

        # calculate and return current eta_p and yp
        eta_p = self.decision_function(X)
        yp = self._link(eta_p, inverse=True)

        return yp, eta_p

    def predict(self, X: np.ndarray, model_index: Optional[int] = None) -> np.ndarray:
        """
        Compute model predictions in the original target space.

        Parameters
        ----------
        X: np.ndarray
            The model matrix. If `init_type` was set as "residuals", then the model
            scores for which to calculate residuals should form the last column of the
            input matrix.

        model_index: int, optional (default=None)
            If None, then return the full model prediction as the sum of all models in
            the ensemble plus the initial model prediction. If an int, then return only
            the predictions from models up to `model_index` (i.e., [:model_index]).

        Returns
        -------
        predictions: np.ndarray
        """
        eta_p = self.decision_function(X, model_index)
        return self._link(eta_p, inverse=True)

    def prediction_history(self, X: np.ndarray, links: bool = False) -> np.ndarray:
        """
        Compute a prediction history.

        This will compute a matrix of predictions with each column corresponding to the
        predictions up to the underlying ensemble at that column index.

        Parameters
        ----------
        X: np.ndarray
            The model matrix. If `init_type` was set as "residuals", then the model
            scores for which to calculate residuals should form the last column of the
            input matrix.

        links: bool (default=False)
            If true, then return the links of the prediction history. Otherwise, return
            the non-transformed predictions.
        """
        model_links = self.get_model_links(X)
        lr_array = np.hstack([[1], [tup[1] for tup in self._model_list]]).reshape(
            (1, -1)
        )
        model_links *= lr_array
        link_history = np.cumsum(model_links, axis=1)

        if links is True:
            return link_history

        return self._link(link_history, inverse=True)

    def reset_model(self) -> None:
        """
        Reset the model fit status to False.

        This will cause the model to reinitialize if the fit method is called after
        the reset_model method.
        """
        self._is_fit = False

    def _compute_beta(
        self, yt: np.ndarray, eta_p: np.ndarray, next_model_preds: np.ndarray
    ) -> float:
        """
        Private method - learning rate switch statement.

        Redirect learning rate calcs as per the user-specified learning rate method.
        """
        if self.step_type == "decaying":
            return self._line_search_decaying(yt, eta_p, next_model_preds)

        if self.step_type == "best":
            return self._line_search_best(yt, eta_p, next_model_preds)

        if self.step_type == "constant":
            return self._beta

        raise AttributeError("init arg:<step_type> is mis-specified")

    def _line_search_best(
        self, yt: np.ndarray, eta_p: np.ndarray, next_model_preds: np.ndarray
    ) -> float:
        """Private method - determine the learning rate for the "best" `step_type`."""
        self._beta = self.alpha
        beta = self._line_search_decaying(yt, eta_p, next_model_preds)
        return beta

    def _line_search_decaying(
        self, yt: np.ndarray, eta_p: np.ndarray, next_model_preds: np.ndarray
    ) -> float:
        """Private method - determine the learning rate for the "decaying" `step_type`."""
        beta0 = self._beta
        etas = eta_p + next_model_preds * beta0
        preds = self._link(etas, inverse=True)
        loss0 = self._loss(yt, preds).mean()
        while True:
            beta = beta0 * self.step_decay_factor
            etas = eta_p + next_model_preds * beta
            preds = self._link(etas, inverse=True)
            loss = self._loss(yt, preds).mean()
            if loss < loss0:
                loss0 = loss
                beta0 = beta
            else:
                self._beta = beta0
                break
        return beta0

    def _track_loss(
        self,
        yp_train: np.ndarray,
        yp_val: Optional[np.ndarray],
        model_data: ModelDataSets,
    ) -> None:
        """
        Private method - track model loss history.

        Implements the logic to track training and validation loss during the model
        fitting process.
        """
        tloss = np.sum(
            self._loss(model_data.yt_train, yp_train) * model_data.weights_train
        ) / np.sum(model_data.weights_train)

        if model_data.has_validation_set():
            vloss = np.sum(
                self._loss(model_data.yt_val, yp_val)
                * model_data.weights_val
                / np.sum(model_data.weights_val)
            )
        else:
            vloss = np.nan
        self._loss_list.append((tloss, vloss))

    def _stop_model(self, model_data: ModelDataSets) -> bool:
        """Private method that contains the logic for the stopping criteria."""
        # training loss condition
        if self.tol is not None:
            tloss = self._loss_list[-2][0] - self._loss_list[-1][0]
            if tloss < self.tol:
                return True

        # validation loss condition
        if model_data.has_validation_set():
            if (
                self.get_iterations() > self.validation_iter_stop
                and self._loss_list[-self.validation_iter_stop - 1][1]
                < self._loss_list[-1][1]
            ):
                return True

        return False

    class InitialModel:
        """InitialModel class implementation - internal to BoostedModel."""

        def __init__(self, link: BaseLink, init_type: str = "mean") -> None:
            """
            Class initializer.

            Parameters
            ----------
            link: BaseLink
                Link function to use for initialization.

            init_type: str (default="mean")
                One of "mean", "residuals", or "zero".
            """
            self._link = link  # type: BaseLink
            self._init_type = init_type  # type: str
            self._value = 0.0  # type: float

        def fit(
            self, X: np.ndarray, yt: np.ndarray, weights: Optional[np.ndarray] = None
        ) -> Model:
            """
            Fit the InitialModel object.

            Parameters
            ----------
            X: np.ndarray
                The model matrix. If `init_type` was set as "residuals", then the model
                scores for which to calculate residuals should form the last column of the
                input matrix.

            yt: np.ndarray
                Observed target values.

            weights: np.ndarray, optional (default=None)
                Sample weights (by observation) to use for fitting. Should be positive.
                Observations with higher weights will affect the model fit more. If 'None',
                then all weights will be equal (1.0).

            Returns
            -------
            self
            """
            weights = np.ones_like(yt) if weights is None else weights
            if self._init_type in ["zero", "residuals"]:
                self._value = 0.0
            elif self._init_type == "mean":
                value = np.sum(yt * weights) / np.sum(weights)
                self._value = self._link(value)
            else:
                raise AttributeError("init arg:<init_type> is mis-specified")

            return self

        def predict(self, X: np.ndarray) -> np.ndarray:
            """
            Compute InitialModel predictions.

            Parameters
            ----------
            X: np.ndarray
                The model matrix. If `init_type` was set as "residuals", then the model
                scores for which to calculate residuals should form the last column of the
                input matrix.

            Returns
            -------
            predictions: np.ndarray
            """
            if self._init_type == "residuals":
                return self._link(X[:, -1])

            if self._init_type == "offset":
                return self._value + X[:, -1]

            return np.ones(X.shape[0]) * self._value



"""Loss function abstract base class."""

# author: Benjamin Cross
# email: btcross26@yahoo.com
# created: 2019-08-26

from abc import ABC, abstractmethod

import numpy as np


class BaseLoss(ABC):
    """Base class for loss functions."""

    def __call__(self, yt: np.ndarray, yp: np.ndarray) -> np.ndarray:
        """Call the instance object to calculate the loss function."""
        return self._loss(yt, yp)

    @abstractmethod
    def _loss(self, yt: np.ndarray, yp: np.ndarray) -> np.ndarray:
        """Calculate the per-observation loss as a function of `yt` and `yp`."""
        ...

    @abstractmethod
    def dldyp(self, yt: np.ndarray, yp: np.ndarray) -> np.ndarray:
        """Calculate the first derivative of the loss with respect to `yp`."""
        pass

    @abstractmethod
    def d2ldyp2(self, yt: np.ndarray, yp: np.ndarray) -> np.ndarray:
        """Calculate the second derivative of the loss with respect to `yp`."""
        pass



"""Student's t function implementation."""

# author: Benjamin Cross
# email: btcross26@yahoo.com
# created: 2019-08-26

from typing import Optional

import numpy as np

from .base_class import BaseLoss


class StudentTLoss(BaseLoss):
    """Student's t loss function class."""

    def __init__(self, scale: Optional[float] = None, dof: int = 2):
        """
        Class initializer.

        Extends BaseLoss.__init__.

        Parameters
        ----------
        scale: float, optional (default=None)
            The `scale` of the Student's t distribution to use. If set to None, the
            default, then taken as (`dof` + 1.0) / 2.0.

        dof: int
            The number of degrees of freedom for the Student's t distribution to use.
        """
        super().__init__()
        self.nu_ = dof
        self.scale_ = (dof + 1.0) / 2.0 if scale is None else scale

    def _loss(self, yt: np.ndarray, yp: np.ndarray) -> np.ndarray:
        """
        Calculate the per-observation loss as a function of `yt` and `yp`.

        Overrides BaseLoss._loss.
        """
        return self.scale_ * np.log(1.0 + (yt - yp) ** 2 / self.nu_)

    def dldyp(self, yt: np.ndarray, yp: np.ndarray) -> np.ndarray:
        """
        Calculate the first derivative of the loss with respect to `yp`.

        Overrides BaseLoss.dldyp.
        """
        return -2.0 * self.scale_ * (yt - yp) / (self.nu_ + (yt - yp) ** 2)

    def d2ldyp2(self, yt: np.ndarray, yp: np.ndarray) -> np.ndarray:
        """
        Calculate the second derivative of the loss with respect to `yp`.

        Overrides BaseLoss.d2ldyp2.
        """
        left_term_du = -2.0 * self.scale_ / (self.nu_ + (yt - yp) ** 2)
        right_term_du = (
            4.0 * self.scale_ * ((yt - yp) ** 2) / ((self.nu_ + (yt - yp) ** 2) ** 2)
        )
        return -(left_term_du + right_term_du)



"""
Unit tests for single-parameter loss function implementations
"""

# author: Benjamin Cross
# email: btcross26@yahoo.com
# created: 2019-12-17


import logging
import warnings

import numpy as np
import pytest

from genestboost.utils import d1_central_difference, d2_central_difference

from .loss_test_list import LOSS_TESTS

LOGGER = logging.getLogger(__name__)


# constants
TEST_NAMES = list(map(lambda x: x[0], LOSS_TESTS))


# generic class to test loss functions
@pytest.mark.parametrize(
    "test_name,loss,loss_test_values,tol", LOSS_TESTS, ids=TEST_NAMES
)
class TestLossFunction:
    """
    Generic unit tests for loss functions
    """

    def test_loss_function(self, test_name, loss, loss_test_values, tol):
        """
        Test loss calculations
        """
        # GIVEN some true (yt, yp, error) tuples of loss function values
        yt = loss_test_values[:, 0]
        yp = loss_test_values[:, 1]
        error = loss_test_values[:, 2]

        # WHEN the loss function is called on yt, yp
        calculated_values = loss(yt, yp)

        # THEN the correct values of error should be returned
        np.testing.assert_allclose(calculated_values, error, atol=1e-4, rtol=1e-5)

    def test_d1_loss(self, test_name, loss, loss_test_values, tol):
        """
        Test loss first derivative calculations
        """
        # GIVEN some yt, yp locations to calculate link derivatives
        yt = loss_test_values[:, 0]
        yp = loss_test_values[:, 1]

        # WHEN the first derivative is approximated using central differences
        def func(y):
            return loss(yt, y)

        calculated_values = d1_central_difference(func, yp, h=1e-8)

        # THEN the calculated values of the derivative should be close to the
        # true implemented computed values
        expected_values = loss.dldyp(yt, yp)
        np.testing.assert_allclose(
            calculated_values, expected_values, atol=tol[0], rtol=tol[1]
        )

    def test_d2_loss(self, test_name, loss, loss_test_values, tol):
        """
        Test loss second derivative calculations
        """
        # GIVEN some yt, yp locations to calculate link derivatives
        yt = loss_test_values[:, 0]
        yp = loss_test_values[:, 1]

        # WHEN the second derivative is approximated using central differences
        def func(y):
            return loss(yt, y)

        calculated_values = d2_central_difference(func, yp, h=1e-6)

        # THEN the calculated values of the derivative should be close to the
        # true implemented computed values
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            expected_values = loss.d2ldyp2(yt, yp)
        np.testing.assert_allclose(
            calculated_values, expected_values, atol=tol[0], rtol=tol[1]
        )



"""genestboost.utils module."""

# author: Benjamin Cross
# email: btcross26@yahoo.com
# created: 2019-12-16

from .gradients import d1_central_difference, d2_central_difference

__all__ = ["d1_central_difference", "d2_central_difference"]



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config
# apidoc command: sphinx-apidoc \
#                   -d 2 \
#                   -f \
#                   -e \
#                   -M \
#                   -H genestboost \
#                   -A Benjamin\ Cross \
#                   -o ./ \
#                   ../ \
#                   ../setup*

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#


import os
import re
import sys
from pathlib import Path

from recommonmark.parser import CommonMarkParser

sys.path.insert(0, os.path.abspath("../.."))


# HELPER functions
def skip(app, what, name, obj, would_skip, options):
    """Document __init__ class method."""
    if name == "__init__":
        return False
    return would_skip


def setup(app):
    """Implement the above skip method."""
    app.connect("autodoc-skip-member", skip)


def dedupe_submodules():
    """
    Clear out duplicate submodules.
    """
    dedupe_list = [
        "genestboost.link_functions.rst",
        "genestboost.loss_functions.rst",
        "genestboost.utils.rst",
        "genestboost.weak_learners.rst",
        "genestboost.rst",
    ]

    for fn in dedupe_list:
        with open(fn, "rt") as f:
            generated_content = f.read()

        new_content = re.sub(
            "Submodules.*\nModule contents\n-*\n",
            "",
            generated_content,
            flags=re.DOTALL,
        )

        with open(fn, "wt") as f:
            f.write(new_content)


# -- Project information -----------------------------------------------------

REPO = Path(__file__).absolute().parent.parent.parent  # repo path from conf.py
PACKAGE_INIT = REPO / "genestboost" / "__init__.py"
project = "genestboost"
copyright = "2021, Benjamin Cross"
author = "Benjamin Cross"
version = re.match(
    r".*__version__ *= *\"([\w\.-]+)*?\".*", PACKAGE_INIT.read_text(), re.DOTALL
).group(1)


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.viewcode",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.githubpages",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",
    "sphinx.ext.intersphinx",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# source parsers
source_parsers = {
    ".md": CommonMarkParser,
}
source_suffix = [".rst", ".md"]

# masterdoc
master_doc = "index"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

html_theme_options = {
    "titles_only": False,
    "includehidden": True,
    "navigation_depth": 4,
    "canonical_url": "",
    "analytics_id": "",
    "logo_only": False,
    "sticky_navigation": True,
    "collapse_navigation": True,
    "display_version": True,
    "prev_next_buttons_location": "bottom",
    "style_external_links": True,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# sidebar templates
html_sidebars = {
    "**": ["about.html", "navigation.html", "relations.html", "searchbox.html",]
}

# -- Extension configuration -------------------------------------------------
autodoc_typehints = "none"
typehints_document_rtype = False

# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

dedupe_submodules()



"""Identity link function implementation."""

# author: Benjamin Cross
# email: btcross26@yahoo.com
# created: 2019-08-26


import numpy as np

from .base_class import BaseLink


class IdentityLink(BaseLink):
    """Implementation of the IdentityLink function."""

    def _link(self, y: np.ndarray) -> np.ndarray:
        """
        Get the link, eta, as a function of y.

        Overrides BaseLink._link.
        """
        return 1.0 * y

    def _inverse_link(self, eta: np.ndarray) -> np.ndarray:
        """
        Get the target, y, as a function of the link, `eta`.

        Overrides BaseLink._inverse_link.
        """
        return 1.0 * eta

    def dydeta(self, y: np.ndarray) -> np.ndarray:
        """
        Get the derivative of `y` with respect to the link as a function of y.

        Overrides BaseLink.dydeta.
        """
        return np.ones(y.shape)

    def d2ydeta2(self, y: np.ndarray) -> np.ndarray:
        """
        Get the second derivative of `y` with respect to the link as a function of y.

        Overrides BaseLink.d2ydeta2.
        """
        return np.zeros(y.shape)



"""Quantile regression loss function implementation."""

# author: Benjamin Cross
# email: btcross26@yahoo.com
# created: 2019-08-27

import warnings

import numpy as np

from .base_class import BaseLoss


class QuantileLoss(BaseLoss):
    """Quantile regression loss function class."""

    def __init__(self, p: float):
        """
        Class initializer.

        Extends BaseLoss.__init__.

        Parameters
        ----------
        p: float
            Value in the interval (0, 1) representing the quantile to regress on.
        """
        super().__init__()
        self.p_ = p

    def _loss(self, yt: np.ndarray, yp: np.ndarray) -> np.ndarray:
        """
        Calculate the per-observation loss as a function of `yt` and `yp`.

        Overrides BaseLoss._loss.
        """
        multiplier = np.where(yt - yp < 0, 1.0 - self.p_, self.p_)
        return np.abs(yt - yp) * multiplier

    def dldyp(self, yt: np.ndarray, yp: np.ndarray) -> np.ndarray:
        """
        Calculate the first derivative of the loss with respect to `yp`.

        Overrides BaseLoss.dldyp.
        """
        return np.where(yt - yp < 0, 1.0 - self.p_, -self.p_)

    def d2ldyp2(self, yt: np.ndarray, yp: np.ndarray) -> np.ndarray:
        """
        Calculate the second derivative of the loss with respect to `yp`.

        Overrides BaseLoss.d2ldyp2.
        """
        warnings.warn(
            "second derivative of quantile value loss with respect to yp is zero"
        )
        return np.zeros(yp.shape)



"""Package setup file."""

import re
from pathlib import Path

from setuptools import find_packages, setup

# Core package components and metadata

NAME = "genestboost"
EMAIL = "btcross26@yahoo.com"
PACKAGES = find_packages()
KEYWORDS = [""]
DESCRIPTION = "General boosting framework for any regression estimator"
LONG = """
:code:`genestboost` is an ML boosting library that separates the modeling algorithm from
the boosting algorithm. The result is that you can boost any generic regression
model, not just trees. Build a forward-thinking (forward-propagating) neural network if
you wish, or build an ensemble of support vector machines if you would so desire. Mix
and match link and loss functions at will.
"""

PROJECT_URLS = {
    "Documentation": ("https://btcross26.github.io/genestboost/build/html/index.html"),
    "Bug Tracker": "https://github.com/btcross26/genestboost/issues",
    "Source Code": "https://github.com/btcross26/genestboost",
}

CLASSIFIERS = [
    "Natural Language :: English",
    "Operating System :: OS Independent",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
]

INSTALL_REQUIRES = [
    "numpy==1.*,>=1.18.5",
    "scipy==1.*,>=1.4.1",
]

EXTRAS_REQUIRE = {
    "docs": ["sphinx", "sphinx_rtd_theme", "sphinx-autodoc-typehints", "recommonmark"],
    "tests": ["coverage", "mypy", "pytest", "pytest-cov", "toml"],
    "qa": [
        "pre-commit",
        "black",
        "mypy",
        "tox",
        "check-manifest",
        "flake8",
        "flake8-docstrings",
    ],
    "build": ["twine", "wheel"],
    "notebooks": [
        "jupyter",
        "ipykernel",
        "matplotlib==3.*,>=3.2.1",
        "pandas==1.*,>=1.0.4",
        "scikit-learn==0.*,>=0.22",
        "patsy==0.5.1",
    ],
}

EXTRAS_REQUIRE["dev"] = (
    EXTRAS_REQUIRE["tests"]
    + EXTRAS_REQUIRE["docs"]
    + EXTRAS_REQUIRE["qa"]
    + EXTRAS_REQUIRE["build"]
    + EXTRAS_REQUIRE["notebooks"]
)


HERE = Path(__file__).absolute().parent
PACKAGE_INIT = HERE / "genestboost" / "__init__.py"
VERSION = re.match(
    r".*__version__ *= *\"([\w\.-]+)*?\".*", PACKAGE_INIT.read_text(), re.DOTALL
).group(1)
URL = PROJECT_URLS["Source Code"]
AUTHORS = (HERE / "AUTHORS").read_text().split("\n")
LICENSE = "BSD 3-Clause"


def install_pkg():
    """Configure the setup for the package."""
    setup(
        name=NAME,
        version=VERSION,
        description=DESCRIPTION,
        long_description=LONG,
        long_description_content_type="text/x-rst",
        url=URL,
        project_urls=PROJECT_URLS,
        author=AUTHORS[0],
        author_email=EMAIL,
        maintainer=AUTHORS[0],
        license=LICENSE,
        python_requires=">=3.7.0,<3.10",
        packages=PACKAGES,
        install_requires=INSTALL_REQUIRES,
        classifiers=CLASSIFIERS,
        extras_require=EXTRAS_REQUIRE,
        include_package_data=True,
        zip_safe=False,
    )


if __name__ == "__main__":
    install_pkg()



"""Least squares loss function implementation."""

# author: Benjamin Cross
# email: btcross26@yahoo.com
# created: 2019-08-26

import numpy as np

from .base_class import BaseLoss


class LeastSquaresLoss(BaseLoss):
    """Least squares loss function class."""

    def _loss(self, yt: np.ndarray, yp: np.ndarray) -> np.ndarray:
        """
        Calculate the per-observation loss as a function of `yt` and `yp`.

        Overrides BaseLoss._loss.
        """
        return 0.5 * (yt - yp) * (yt - yp)

    def dldyp(self, yt: np.ndarray, yp: np.ndarray) -> np.ndarray:
        """
        Calculate the first derivative of the loss with respect to `yp`.

        Overrides BaseLoss.dldyp.
        """
        return yp - yt

    def d2ldyp2(self, yt: np.ndarray, yp: np.ndarray) -> np.ndarray:
        """
        Calculate the second derivative of the loss with respect to `yp`.

        Overrides BaseLoss.d2ldyp2.
        """
        return np.ones(yp.shape)
